// These functions read in .ini (initialize) and .par (parameter) files as well as print out frames and the .sta (statistics) files

#include <string.h>
#include <stdio.h>
#include <iostream>
#include "LN.h"

int LymphNode::readIniFile(char* filename) 
{
  char s[1200];
  char spar[100];
  char sval[1200]; // if you have a problem in this function, it is probably because this array is too small
  int i = 0;
  int j = 0;
  int n = 0;

  ifstream inifile(filename);
  
  if (inifile == NULL){
    return -1;
  }
  //cout <<"Reading the .ini input file" << endl;
  while (!inifile.eof()) {   
    inifile.getline(s, 1200); // this number needs to be the size of s and the size of sval
    if (s[0] == '/' || s[0] == 0){
      continue;
    }

    i = 0;
    while (s[i] != 0){
      if (s[i]=='='){
	break;
      }
      i++;
    }
    
    if (s[i] == 0){
      return -1;  // No = in string
    }

    j = i - 1;
    while(s[j] == ' '){
      j--;
    }

    strncpy(spar, s, j + 1);
    spar[j + 1] = 0;
    
    j = i + 1;
    while(s[j] == ' '){
      j++;
    }
    strcpy(sval, &s[j]);

    if (strcmp(spar, "SOURCES") == 0){
      if (sval[0] != '{'){
	return -1;
      }
      if (sval[1] == '}') {
	sources = NULL;
	NSOURCES = 0;
	continue;
      }
      
      j = 0;
      n = 0;
      while (sval[j] != 0) {
	if (sval[j] == ','){
	  n++;
	}
	j++;
      }
      try
      {
    	  sources = new int[n+1];
      }
      catch (bad_alloc error)
      {
          cout << error.what() << endl;
          cout<< "Memory could not be allocated to new sources in readIniFile() " << endl;
          cout << "Exiting...." << endl;
          exit(1);
      }
      sources[0] = atoi(&sval[1]);
      
      j = 1;
      n = 1;
      while (sval[j] != 0) {
	if (sval[j] == ',') {
	  sources[n] = atoi(&sval[j + 1]);
	  n++;
	} 
	j++;
      }
      NSOURCES = (n + 1) / 2;
    }

    if (strcmp(spar, "IDC_AFFERENT") == 0) {
      if (sval[0] != '{'){
	return -1;
      }
      if (sval[1] == '}') {
	afferent = NULL;
	NIDC_AFFERENT = 0;
	continue;
      }

      j = 0;
      n = 0;
      while (sval[j] != 0) {
	if (sval[j] == ','){
	  n++;
	}
	j++;
      }
      try
      {
    	  idc_afferent = new int[n + 1];
      }
      catch (bad_alloc error)
      {
          cout << error.what() << endl;
          cout<< "Memory could not be allocated to idc_afferent in readIniFile() " << endl;
          cout << "Exiting...." << endl;
          exit(1);
      }
      idc_afferent[0] = atoi(&sval[1]);
      
      j = 1;
      n = 1;
      while (sval[j] != 0) {
	if (sval[j] == ',') {
	  idc_afferent[n] = atoi(&sval[j + 1]);
	  n++;
	}
	j++;
      }
      NIDC_AFFERENT = (n + 1) / 2;
    }


    if (strcmp(spar, "AFFERENT") == 0) {
      if (sval[0] != '{'){
	return -1;
      }
      if (sval[1] == '}') {
	afferent = NULL;
	NAFFERENT = 0;
	continue;
      }
      
      j = 0;
      n = 0;
      while (sval[j] != 0) {
	if (sval[j] == ','){
	  n++;
	}
	j++;
      }
      try
      {
    	  afferent = new int[n + 1];
      }
      catch (bad_alloc error)
      {
          cout << error.what() << endl;
          cout<< "Memory could not be allocated to new afferent in readIniFile() " << endl;
          cout << "Exiting...." << endl;
          exit(1);
      }
      afferent[0] = atoi(&sval[1]);
      
      j = 1;
      n = 1;
      while (sval[j] != 0) {
	if (sval[j] == ',') {
	  afferent[n] = atoi(&sval[j + 1]);
	  n++;
	} 
	j++;
      }      
      NAFFERENT = (n + 1) / 2;
    }
     
    if (strcmp(spar, "EFFERENT") == 0) {
      if (sval[0] != '{'){
	return -1;
      }
      if (sval[1] == '}') {
	efferent = NULL;
	NEFFERENT = 0;
	continue;
      }
      j = 0;
      n = 0;
      while (sval[j] != 0) {
	if (sval[j] == ','){
	  n++;
	}
	j++;
      }
      try
      {
    	  efferent = new int[n + 1];
      }
      catch (bad_alloc error)
      {
          cout << error.what() << endl;
          cout<< "Memory could not be allocated to new efferent in readIniFile() " << endl;
          cout << "Exiting...." << endl;
          exit(1);
      }
      efferent[0] = atoi(&sval[1]);
      
      j = 1;
      n = 1;
      while (sval[j] != 0) {
	if (sval[j] == ',') {
	  efferent[n] = atoi(&sval[j + 1]);
	  n++;
	} 
	j++;
      } 
      NEFFERENT = (n + 1) / 2;
    }
  }
  return 0;
}

int LymphNode::readParFile(char* filename) 
{
  //cout << "Reading .par file " << endl;
  char s[100];
  char spar[100];
  char sval[100];
  int i = 0;
  int j = 0;

  ifstream inifile(filename);
  
  if (inifile == NULL) return -1;
  //cout << "Reading .par file " << endl;
  while (!inifile.eof()) {
    inifile.getline(s, 100);  // read one line
    if (s[0] == '/' || s[0] == 0) continue;  // if line begins with / or 0 , ignore it
    
    i = 0;
    while (s[i] != 0){   // while ith character in not 0 keep reading
      if (s[i] == '=') break; // if ith character is = , stop reading
      i++;   
    }
    
    if (s[i] == 0) return -1;  // No = in string
    j = i - 1;
    while(s[j] == ' ') j--; // remove whitespace 
    
    strncpy(spar, s, j + 1); // copy the string into spar
    spar[j + 1] = 0; // end of spar
    j = i + 1;
    while(s[j] == ' ') j++;  // skip whitespace
    strcpy(sval, &s[j]); // read the rest of s into sval
    
    if (strcmp(spar, "NDAYS") == 0) 
      p.setNDAYS(atoi(sval));
    else if (strcmp(spar, "INITIAL_rHELPERT_CELLS") == 0) 
      p.setINITIAL_rHELPERT_CELLS(atoi(sval)); 
    else if (strcmp(spar, "INITIAL_aHELPERT_CELLS") == 0) 
      p.setINITIAL_aHELPERT_CELLS(atoi(sval)); 
    else if (strcmp(spar, "INITIAL_eHELPERT_CELLS") == 0)  
      p.setINITIAL_eHELPERT_CELLS(atoi(sval));  
    else if (strcmp(spar, "INITIAL_rKILLERT_CELLS") == 0)
      p.setINITIAL_rKILLERT_CELLS(atoi(sval)); 
    else if (strcmp(spar, "INITIAL_aKILLERT_CELLS") == 0)  
      p.setINITIAL_aKILLERT_CELLS(atoi(sval));  
    else if (strcmp(spar, "INITIAL_mKILLERT_CELLS") == 0)  
      p.setINITIAL_mKILLERT_CELLS(atoi(sval));  
    else if (strcmp(spar,"INITIAL_DENDRITIC_CELLS")==0) 
      p.setINITIAL_DENDRITIC_CELLS(atoi(sval)); 
    else if (strcmp(spar,"PEAK_DEN_CELLS")==0)
                p.setPEAK_DEN_CELLS(atoi(sval));
    
    else if (strcmp(spar,"A")==0)
         p.setA(atoi(sval));
    else if (strcmp(spar,"B")==0)
         p.setB(atoi(sval));
    else if (strcmp(spar,"C")==0)
         p.setC(atoi(sval));
    else if (strcmp(spar,"D")==0)
         p.setD(atoi(sval));

    //else if (strcmp(spar,"INITIAL_mDENDRITIC_CELLS")==0)
    //  p.setINITIAL_mDENDRITIC_CELLS(atoi(sval)); 
    //else if (strcmp(spar,"INITIAL_lDENDRITIC_CELLS")==0) 
    //  p.setINITIAL_lDENDRITIC_CELLS(atoi(sval)); 
    
    /*else if (strcmp(spar,"CYTOKINES_SECRETED_BY_HCELL")==0) 
      CYTOKINES_SECRETED_BY_HCELL = atoi(sval); 
    else if (strcmp(spar,"CYTOKINES_SECRETED_BY_KCELL")==0)
      CYTOKINES_SECRETED_BY_KCELL = atoi(sval);
    else if (strcmp(spar,"CHEMOKINES_SECRETED_BY_DC")==0)  
    CHEMOKINES_SECRETED_BY_DC = atoi(sval);  
    else if (strcmp(spar,"IL2_THRESHOLD")==0)  
    IL2_THRESHOLD = atof(sval); */ 

    
    else if (strcmp(spar, "PROB_eHELPER_LICENSES_MDC") == 0)
      p.setPROB_eHELPER_LICENSES_MDC(atof(sval));
    
    //else if (strcmp(spar, "PROB_rHELPER_LICENSES_MDC") == 0)
    //  p.setPROB_rHELPER_LICENSES_MDC(atof(sval));
    
    else if (strcmp(spar,"pMHC_HALFLIFE")==0)
      p.setpMHC_HALFLIFE(atoi(sval));
    
    else if (strcmp(spar,"UNBINDING_THRESHOLD")==0)
          p.setUNBINDING_THRESHOLD(atoi(sval));
    
    else if (strcmp(spar,"BINDING_THRESHOLD")==0)
      p.setBINDING_THRESHOLD(atoi(sval));
    
    else if (strcmp(spar,"IDC_MDC_BOUNDARY")==0)
          p.setIDC_MDC_BOUNDARY(atoi(sval));
    
    else if (strcmp(spar,"DECAY_TCELL_CONTACT_HISTORY")==0)
              p.setDECAY_TCELL_BINDING_HISTORY(atoi(sval));
    
    else if (strcmp(spar,"DECAY_TCELL_CONTACT_HISTORY")==0)
          p.setDECAY_TCELL_BINDING_HISTORY(atoi(sval));
    
    else if (strcmp(spar,"pMHC_FOR_MEDIAN_BINDING_PROB_HCELL")==0)
      p.setpMHC_FOR_MEDIAN_BINDING_PROB_HCELL(atoi(sval));
    
    else if (strcmp(spar,"SHAPE_BINDING_PROB_CURVE_HCELL")==0)
      p.setSHAPE_BINDING_PROB_CURVE_HCELL(atoi(sval));
    
    else if (strcmp(spar,"TIME_FOR_MEDIAN_ACTIVATION_PROB_HCELL")==0)
      p.setTIME_FOR_MEDIAN_ACTIVATION_PROB_HCELL(atoi(sval));
    
    else if (strcmp(spar,"SHAPE_ACTIVATION_PROB_CURVE_HCELL")==0)
             p.setSHAPE_ACTIVATION_PROB_CURVE_HCELL(atoi(sval));
    
    else if (strcmp(spar,"pMHC_FOR_MEDIAN_BINDING_PROB_KCELL")==0)
      p.setpMHC_FOR_MEDIAN_BINDING_PROB_KCELL(atoi(sval));
    
    else if (strcmp(spar,"SHAPE_BINDING_PROB_CURVE_KCELL")==0)
      p.setSHAPE_BINDING_PROB_CURVE_KCELL(atoi(sval));
    
    else if (strcmp(spar,"TIME_FOR_MEDIAN_ACTIVATION_PROB_KCELL")==0)
      p.setTIME_FOR_MEDIAN_ACTIVATION_PROB_KCELL(atoi(sval));
    
    else if (strcmp(spar,"SHAPE_ACTIVATION_PROB_CURVE_KCELL")==0)
             p.setSHAPE_ACTIVATION_PROB_CURVE_KCELL(atoi(sval));

    
    else if (strcmp(spar, "PROB_aK_BECOMES_mK") == 0)
      p.setPROB_aK_BECOMES_mK(atof(sval));
    
    //else if (strcmp(spar, "PROB_LDC_MATURES_DC") == 0)
    //  p.setPROB_LDC_MATURES_DC(atof(sval));
    
    //else if (strcmp(spar, "PROB_MDC_MATURES_IDC") == 0)
    //  p.setPROB_MDC_MATURES_IDC(atof(sval));
    
    else if (strcmp(spar, "PROB_H_ENHANCED_RECRUITMENT") == 0)
      p.setPROB_H_ENHANCED_RECRUITMENT(atof(sval));
    else if (strcmp(spar, "PROB_K_ENHANCED_RECRUITMENT") == 0)
      p.setPROB_K_ENHANCED_RECRUITMENT(atof(sval));
    else if (strcmp(spar, "PROB_DC_ENHANCED_RECRUITMENT") == 0)
      p.setPROB_DC_ENHANCED_RECRUITMENT(atof(sval));
    
    //else if (strcmp(spar, "PROB_BINDING_PRIMED_T4_WITH_IDC") == 0)
    //  p.setPROB_BINDING_PRIMED_T4_WITH_IDC(atof(sval));  
    
    else if (strcmp(spar, "PROB_BINDING_T4_WITH_DC") == 0)
      p.setPROB_BINDING_T4_WITH_DC(atof(sval));
    else if (strcmp(spar, "PROB_BINDING_T8_WITH_LDC") == 0)
      p.setPROB_BINDING_T8_WITH_LDC(atof(sval));
    else if (strcmp(spar,"PROB_DC_ACTIVATES_T4")==0) 
      p.setPROB_DC_ACTIVATES_T4(atof(sval));
    else if (strcmp(spar,"PROB_LDC_ACTIVATES_T8")==0)
      p.setPROB_LDC_ACTIVATES_T8(atof(sval));
    
    // public attributes
    else if (strcmp(spar,"PROB_H_RECRUITMENT")==0)
      p.PROB_H_RECRUITMENT = atof(sval);
    else if (strcmp(spar,"PROB_K_RECRUITMENT")==0)
      p.PROB_K_RECRUITMENT = atof(sval);
    else if (strcmp(spar,"PROB_D_RECRUITMENT")==0)
      p.PROB_D_RECRUITMENT = atof(sval);
    
    /*else if (strcmp(spar,"CHEMOKINES_THRESHOLD_FOR_ENHANCED_H_RECRUITMENT")==0)
      CHEMOKINES_THRESHOLD_FOR_ENHANCED_H_RECRUITMENT = atof(sval);
    else if (strcmp(spar,"CHEMOKINES_THRESHOLD_FOR_ENHANCED_K_RECRUITMENT")==0)
      CHEMOKINES_THRESHOLD_FOR_ENHANCED_K_RECRUITMENT = atof(sval);
    else if (strcmp(spar,"CHEMOKINES_THRESHOLD_FOR_ENHANCED_DC_RECRUITMENT")==0)
    CHEMOKINES_THRESHOLD_FOR_ENHANCED_DC_RECRUITMENT = atof(sval);*/
    
    else if (strcmp(spar,"HCELL_TIME_BEING_ACTIVE")==0)
      p.HCELL_TIME_BEING_ACTIVE = atoi(sval);
    else if (strcmp(spar,"KCELL_TIME_BEING_ACTIVE")==0)
      p.KCELL_TIME_BEING_ACTIVE = atoi(sval);
    else if (strcmp(spar,"DCELL_TIME_BEING_LICENSED")==0)
      	p.DCELL_TIME_BEING_LICENSED = atoi(sval);
    
    else if (strcmp(spar,"aHCELL_DOUBLING_PERIOD")==0)
      p.aHCELL_DOUBLING_PERIOD = atoi(sval); 
    else if (strcmp(spar,"aKCELL_DOUBLING_PERIOD")==0)
      p.aKCELL_DOUBLING_PERIOD = atoi(sval);
    
    else if (strcmp(spar,"MIN_H_CELL_AGE")==0)
      p.MIN_H_CELL_AGE = atoi(sval);
    else if (strcmp(spar,"MAX_H_CELL_AGE")==0) 
      p.MAX_H_CELL_AGE = atoi(sval);
    
    else if (strcmp(spar,"MAX_TIME_H_EFFECTOR")==0)
      p.MAX_TIME_H_EFFECTOR = atoi(sval);
    else if (strcmp(spar,"MAX_TIME_K_EFFECTOR")==0)
      p.MAX_TIME_K_EFFECTOR = atoi(sval);

    else if (strcmp(spar,"MAX_TIME_D_MATURE")==0)
      p.MAX_TIME_D_MATURE = atoi(sval);
 
    else if (strcmp(spar,"MIN_K_CELL_AGE")==0)
      p.MIN_K_CELL_AGE = atoi(sval);
    else if (strcmp(spar,"MAX_K_CELL_AGE")==0)
      p.MAX_K_CELL_AGE = atoi(sval);
 
    else if (strcmp(spar,"MIN_D_CELL_AGE")==0)
      p.MIN_D_CELL_AGE = atoi(sval);
    else if (strcmp(spar,"MAX_D_CELL_AGE")==0)
      p.MAX_D_CELL_AGE = atoi(sval);
 
    //else if (strcmp(spar,"MAX_TIME_NOT_MOVE")==0)
    //  p.MAX_TIME_NOT_MOVE = atoi(sval);
    //else if (strcmp(spar,"MIN_TIME_NOT_MOVE")==0)
    //  p.MIN_TIME_NOT_MOVE = atoi(sval);
    //else if (strcmp(spar,"TIME_REST_T4_LICENSES_MDC")==0)
    //  p.TIME_REST_T4_LICENSES_MDC = atoi(sval);
 
    else if (strcmp(spar,"N_DIVISIONS_T4")==0)
      p.N_DIVISIONS_T4 = atoi(sval);
    else if (strcmp(spar,"N_DIVISIONS_T8")==0)
      p.N_DIVISIONS_T8 = atoi(sval);
 
    else if (strcmp(spar,"N_TCELL_BIND_DC")==0)
      p.N_TCELL_BIND_DC = atoi(sval);
    
    //else if (strcmp(spar, "MAX_TIME_pTC_BOUND_TO_DC") == 0)
    //  p.MAX_TIME_pTC_BOUND_TO_DC = atoi(sval);
    //else if (strcmp(spar, "MIN_TIME_pTC_BOUND_TO_DC") == 0)
    //  p.MIN_TIME_pTC_BOUND_TO_DC = atoi(sval);
    /* else if (strcmp(spar,"PSTRAIGHT")==0)
       PSTRAIGHT = atoi(sval);*/
 
    else if (strcmp(spar,"COGNATE_KCELL")==0) 
      p.COGNATE_KCELL = atoi(sval);
    
    else if (strcmp(spar,"COGNATE_HCELL")==0)
      p.COGNATE_HCELL = atoi(sval);
    
    
    //else if(strcmp(spar, "TOTAL_DEN_CELLS")==0)
    //p.setTOTAL_DEN_CELLS(atoi(sval)); 

    /*else if (strcmp(spar,"PEAK_iDEN_CELLS")==0)
      p.setPEAK_iDEN_CELLS(atoi(sval));*/
    
    /*		else if (strcmp(spar,"SEC_CHEM")==0)
		SEC_CHEM = atoi(sval);*/
    
    /*   		else if (strcmp(spar,"H_SENSE_CHEM")==0)
			H_SENSE_CHEM = atoi(sval);*/
    
    /*		else if (strcmp(spar,"H_PROB_ATTRACTION")==0)
		H_PROB_ATTRACTION = atof(sval);*/
        
    /*		else if (strcmp(spar,"H_TIME_SENSING_CHEM")==0)
		H_TIME_SENSING_CHEM = atoi(sval);*/
    
    /*		else if (strcmp(spar,"H_TIME_NOT_SENSING_CHEM")==0)
		H_TIME_NOT_SENSING_CHEM = atoi(sval);*/
			
    /*		else if (strcmp(spar,"K_SENSE_CHEM")==0)
		K_SENSE_CHEM = atoi(sval);*/
			
    /*		else if (strcmp(spar,"K_PROB_ATTRACTION")==0)
		K_PROB_ATTRACTION = atof(sval);*/
        
    /*		else if (strcmp(spar,"K_TIME_SENSING_CHEM")==0)
		K_TIME_SENSING_CHEM = atoi(sval);*/
 
    /*		else if (strcmp(spar,"K_TIME_NOT_SENSING_CHEM")==0)
		K_TIME_NOT_SENSING_CHEM = atoi(sval);*/
 
  }
  //cout << "Finished reading .par file" << endl;
  return 0;
}

// I don't think this function is used anymore
void LymphNode::save_data( ofstream &stream, char *data,int len) 
{
  for (int i=len-1;i>=0;i--) stream <<data[i];
}

void LymphNode::save_frame() 
{
  ofstream outStream;
  cellpos* ptr = NULL;
  h_cell* hp = NULL;
  k_cell* kp = NULL;
  d_cell* dp = NULL;

  // These variables are all members of the LymphNode class
  NRH = NAH = NEH = NRK = NAK = NEK = NMK = NID = NMD = NLD = 0;
  if (SAVE_FRAME) {
    sprintf(&filename[20], "-%07d.dat", counter);
    outStream.open(filename);
    outStream << "*** helper cells" << endl;
  }
  
  for (int i = LH.size() - 1; i >= 0; i--){
    ptr = LH.elem(i);
    hp = (h_cell *)(ptr->cell_pointer);
    if (SAVE_FRAME) {
      outStream << ptr->row << " " << ptr->col << " " <<hp->id << " " 
		<< hp->birthtime << " " << hp->state << " " 
		<< hp->not_move << " " <<hp->division_counter << " " 
		<< hp->bound_to_dcell << endl;
    }
    
    if (hp->state == H_RESTING || hp->state == H_NOTMOVE_BY_DC) NRH++;
    else if (hp->state == H_ACTIVE) NAH++;
    else NEH++; 
  } 
  
  if (SAVE_FRAME) outStream <<"*** killer cells" << endl;
  
  for (int i = LK.size() - 1; i >= 0; i--){
    ptr = LK.elem(i);
    kp = (k_cell *)(ptr->cell_pointer);
    
    if (SAVE_FRAME) {
      outStream << ptr->row << " " << ptr->col << " " <<kp->id << " " 
		<< kp->birthtime << " " << kp->state << " " 
		<< kp->not_move << " " <<kp->division_counter << " " 
		<< kp->bound_to_dcell << " " << endl;
    }
    
    if (kp->state == K_RESTING || kp->state == K_NOTMOVE_BY_LDC) NRK++;
    else if (kp->state == K_ACTIVE) NAK++;
    else if (kp->state == K_EFFECTOR) NEK++;
    else NMK++; 
  } 
  
  if (SAVE_FRAME)  outStream <<"*** dendritic cells"<<endl;
  
  for (int i = LD.size() - 1; i >= 0; i--){
    ptr = LD.elem(i);
    dp = (d_cell *)(ptr->cell_pointer);
    
    if (SAVE_FRAME) {
      outStream << ptr->row << " " << ptr->col << " " <<dp->id << " " 
		<< dp->birthtime << " " << dp->state << " {";
      for (int j = 0; j < dp->tbind.size(); j++) outStream << dp->tbind[j] << " "; 
      outStream << "}" << endl;
    }
    if (dp->state == D_IMMATURE) NID++;
    else if (dp->state == D_MATURE) NMD++; 
    else  NLD++; 
    
  }  

  if (SAVE_FRAME){
    outStream << "*** afferent" << endl;
    for(int i = 0; i < NAFFERENT; i++){
      outStream << afferent[2 * i] << " " << afferent[2 * i + 1] << endl;
    }
    outStream << "*** efferent v" << endl;
    for(int i = 0; i < NEFFERENT; i++){
      outStream << efferent[2 * i] << " " << efferent[2 * i + 1] << endl;
    }
    outStream.close();
  }
  
  statfile << NRH << "\t" << NAH << "\t" << NEH << "\t" << NRK << "\t" << NAK << "\t" << NEK << "\t"
	   << NMK << "\t" << NID << "\t" << NMD << "\t" << NLD << "\t"
	   << n_t4_cells_leave << "\t" 
	   << n_t4a_cells_leave << "\t"    
	   << n_t4e_cells_leave << "\t"  
	   << n_t8_cells_leave << "\t" 
	   << n_t8a_cells_leave << "\t" 
	   << n_t8e_cells_leave << "\t"  
	   << n_t8m_cells_leave << "\t"  << endl;
}
 
