/* Here are the functions of the dcell (Dendritic cell) class */

#include "LN.h"
#include <stdlib.h>  // for srand(), rand() and RAND_MAX
#include <math.h>
using namespace std;

int nDC_left = 0;
int nDC_created = 0;

// create dendritic cell in M and in the set LD
int LymphNode::create_dcell(int row, int col, int d_time, 
			      int cognate)
{
	int pMHC = 0;
	int rrow = 0;
       int rcol = 0;
	bool foundFreeNeighbor = false;
	//cout << "Site to create DC is [" << row << "][" << col << "]" << endl;
	// Check if this DC can be moved to a neighboring microcompartment so that AL becomes free	
	for(int i = 1; i <= 8; i++) {
    	// look around at our neighbors
    		rrow = i / 3; if (rrow == 2) rrow = -1;
    		rcol = i % 3; if (rcol == 2) rcol = -1; 
		if (row + rrow >= 1 && row + rrow < MROWS - 1 && col + rcol >= 1 && col + rcol < MCOLS - 1)
		{
			if (M[row + rrow][col + rcol].is_empty() ) {
				foundFreeNeighbor = true;
				row = row + rrow;
				col = col + rcol;
				//cout << "Found site to move DC to [" << row << "][" << col << "]" << endl;
				break;
			}
		}
	}	
	 checkRowsAndCols(row, col, "from create_dcell");
	 cellpos cp;
	 cp.row = row;
	 cp.col = col;
	 cp.cell_pointer = NULL;
	 try
	 {
		 cp.cell_pointer = new d_cell();
	 }
	 catch (bad_alloc error)
	 {
		 cout << error.what() << endl;
		 cout<< "Memory could not be allocated to new DC " << endl;
		 cout << "Exiting...." << endl;
		 exit(1);
	 }
	 if (cp.cell_pointer != NULL)
	 {
		d_cell* dptr = (d_cell*) cp.cell_pointer;
		index_dc += 3;
		  //hash = id / 3 - 1;
		cp.cell_pointer->birthtime = d_time;
		cp.cell_pointer->time_in = counter;
		cp.cell_pointer->id = index_dc; //c_time;
		// choose pMHC from uniform distribution
		pMHC = random.Next(0.0, 100.0);
		if (pMHC <= 40.0)
		// choose pMHC from normal distribution
		    pMHC = dptr->normal(50, 20);
		else
		    pMHC = dptr->normal(250, 25.0); 
		if (pMHC < 0 )
			 pMHC = 0;
		if (pMHC > 300 )
			pMHC = 300;
		dptr->initial_pMHC = pMHC;
		if (dptr->initial_pMHC <= p.IDC_MDC_BOUNDARY())
			dptr->state = D_IMMATURE;
		else
			dptr->birthtime = (dptr->normal(20,5)) *  600;
		//cout << "Created dcell, id: " << cp.cell_pointer->id <<", "
		//     << "Initial n_pMHC: " << dptr->initial_pMHC << endl;
		dptr->n_pMHC = dptr->initial_pMHC; // this will be updated every step t_cell
		cp.cell_pointer->next_state = cp.cell_pointer->state;
		cp.cell_pointer->cognate = cognate;
		
		LD.push_back(cp);
		M[row][col].cell[0] = cp.cell_pointer;
		++nDC_created;
		//cout << "**Created DC at AL [" << row << "][" << col << "]** nDC_created:" << nDC_created << endl; 
	 }	
	return nDC_created;	
}

void LymphNode::delete_dcell(int row, int col)
{
  Cell* ptr = M[row][col].cell[0];
  
  M[row][col].cell[0] = NULL;
  for(int i = ((d_cell*)ptr)->tbind.size()-1; i >= 0; i--){
    int tcell = ((d_cell*)ptr)->tbind[i];
    if(tcell % 3 == 2){
      int pos = LH.whereis(tcell);
      if(pos != -1){
	((Tcell*)(LH.elem(pos))->cell_pointer)->bound_to_dcell = 0;
      }else{
	cerr << "hcell: " << tcell << " not removed from tbind." << endl;
      }
    }else if(tcell % 3 == 1){
      int pos = LK.whereis(tcell);
      if(pos != -1){
	((Tcell*)(LK.elem(pos))->cell_pointer)->bound_to_dcell = 0;
      }else{
	cerr << "kcell: " << tcell << " not removed from tbind." << endl;
      }
    }else{
      cerr << "Not a valid tcell id in delete_dcell. Exiting." << endl;
      exit(34);
    }
  }
  //cout << "**DC left the LN** " << ptr->id << "nDC_left: " << ++nDC_left << endl;
  LD.erase(ptr);
   
  
  //delete ptr;
}

void LymphNode::set_dendriticcells_random()
{
  int newrow = 0;
  int newcol = 0;
  int cognate = 0;

  for(int i = 0; i < p.INITIAL_DENDRITIC_CELLS(); i++) {
    newrow = random.Next(1, MROWS-2);
    newcol = random.Next(1, MCOLS-2);
    if(M[newrow][newcol].is_empty()){
      create_dcell(newrow, newcol, -random.Next(p.MIN_D_CELL_AGE, p.MAX_D_CELL_AGE), cognate);
    } // end if
  } // end for
  /*
  for(int i = 0; i < p.INITIAL_mDENDRITIC_CELLS(); i++) {
    newrow = random.Next(1, MROWS-2);
    newcol = random.Next(1, MCOLS-2);
    if(M[newrow][newcol].is_empty()){
      (i % 2 == 0) ? cognate = 1 : cognate = 0;
      create_dcell(newrow, newcol, 0, D_MATURE, cognate);
    } // end if
  } // end for

  for(int i = 0; i < p.INITIAL_lDENDRITIC_CELLS(); i++) {
    newrow = random.Next(1, MROWS-2);
    newcol = random.Next(1, MCOLS-2);
    if(M[newrow][newcol].is_empty()){
      create_dcell(newrow, newcol, 0, D_LICENSED,  cognate);
    } 
  } // end for
  */
}

void LymphNode:: recruit_d_cells() 
{
  for(int i = 0; i < NAFFERENT; i++) 
  {
    if(M[ afferent[2 * i] ][ afferent[2 * i + 1] ].is_empty())
    {
      if (random.Next(0,10000)/100.0 < p.PROB_D_RECRUITMENT) 
      {
    	  //if (random.Next(0,10000)/100.0 < 15) { // the ratio of immature to mature is magic
    		  create_dcell(afferent[2*i], afferent[2*i+1], counter - random.Next(p.MIN_D_CELL_AGE, p.MAX_D_CELL_AGE),  0);
    	  //}else { 
    		  //create_dcell(afferent[2*i], afferent[2*i+1], counter - random.Next(p.MIN_D_CELL_AGE, p.MAX_D_CELL_AGE), D_MATURE, /*cognate*/ 0);
    	  //}
      } // end if prob
    } // end if is_empty
  } // end for
}

/*bool LymphNode::spaceForDcell(int row, int col, int direction)
{
  withinRowsAndCols(row, col);
  return M[row][col].is_empty();
}*/ // end spaceForDcells

double d_cell::normal(const double &mean, const double &std)
{
    static const double pi=3.1415927;
    double num = RAND_MAX;
    double ran_num;
    static const double r_max = num + 1;
    ran_num = std*sqrt(-2*log((rand()+1)/r_max))*sin(2*pi*rand()/r_max)+mean;
    return ran_num;   
}

void d_cell::add_tcell(long hcell_id)
{
  tbind.push_back(hcell_id);
}

void d_cell::remove_tcell(long hcell_id)
{
  for (int j = 0; j < tbind.size(); j++) {
    if (tbind[j] == hcell_id) {
      tbind.erase((vector <int>::iterator)&tbind[j]);
    }
  }
}

void d_cell::add_hcell(long hcell_id)
{
  hcells.push_back(hcell_id);
}

void d_cell::remove_hcell(long hcell_id)
{
  for (int j = 0; j < hcells.size(); j++) {
    if (hcells[j] == hcell_id) {
      hcells.erase((vector <int>::iterator)&hcells[j]);
    }
  }
}

void d_cell::add_kcell(long kcell_id)
{
  kcells.push_back(kcell_id);
}

void d_cell::remove_kcell(long kcell_id)
{ 
  for (int j = 0; j < kcells.size(); j++) {
    if (kcells[j] == kcell_id) {
      kcells.erase((vector <int>::iterator)&kcells[j]);
    }
  }
}

// function to update states of dendritic cells
int LymphNode::next_state_dcell(cellpos *ptr)
{
  d_cell* dcell = (d_cell*)ptr->cell_pointer; 
  int row = ptr->row;
  int col = ptr->col;
  int current_state = dcell->state;
  int rrow = 0;
  int rcol = 0;

  // check the age of the dcell. If its time to die, first they release of 
  // tcells that are bound to it
  if (current_state == D_IMMATURE)
  {
	  if (counter - dcell->birthtime >= p.MAX_D_CELL_AGE )   
	  {       
	   // release tcells
	  	  for (int x = 0; x < dcell->tbind.size(); x++) { 	  
	  		  for (int i = 1; i <= 8; i++) {
	  			  rrow = i / 3; if (rrow == 2) rrow = -1;
	  			  rcol = i % 3; if (rcol == 2) rcol = -1;
	  			  if (row + rrow >= 0 && row + rrow < MROWS 
	  					  && col + rcol >= 0 && col + rcol < MCOLS) {
	  				  for(int j = 0; j < microcompartment::NUM_CELLS; j++){	    
	  					  if (M[row + rrow][col + rcol].cell[j] != NULL) {
	  						  if (dcell->tbind[x] == M[row + rrow][col + rcol].cell[j]->id) {
	  							  ((Tcell*)M[row + rrow][col + rcol].cell[j])->bound_to_dcell = 0;
	  							  M[row + rrow][col + rcol].cell[j]->next_state = H_RESTING;
	  						  }
	  					  }
	  				  }	    
	  			  }
	  		  }    
	  	 } // end for
	  	 return D_DEAD;
	  } // end if counter
	  
	  // check to see if there is a licensed dcell in the neighborhood. If 
	  // so, an immature dcell becomes licenced with probability 
	  // PROB_LDC_LICENSES_IDC	      
	  bool found_LDC = false;
	  d_cell* ldcell;
	  for(int i = 1; i <= 8 && !found_LDC; i++) 
	  {
	  	    rrow = i / 3; if (rrow == 2) rrow = -1;
	  	    rcol = i % 3; if (rcol == 2) rcol = -1;
	        
	  	    if(row + rrow >= 0 && row + rrow < MROWS  && col + rcol >= 0 && col + rcol < MCOLS) 
	  		{
	  			if(M[row + rrow][col + rcol].has_dcell() && ((d_cell*)M[row + rrow][col + rcol].cell[0])->state == D_LICENSED )
	  			{
	  				found_LDC = true;
	  				ldcell = (d_cell*)M[row+rrow][col+rcol].cell[0];
	  			}
	  		}
	  } // end for

	  if (found_LDC) 
	  {
	  		//cout << "dcell n_pMHC: "<< dcell->n_pMHC << ", ";
	  		//cout << " p.BINDING_THRESHOLD(): "<< p.BINDING_THRESHOLD() << ", ";
	  		//cout << "ldcell n_pMHC: " << ldcell->n_pMHC << endl;
	  		dcell->birthtime = counter;
	  		dcell->n_pMHC = (p.BINDING_THRESHOLD() + ldcell->n_pMHC) / 2;
	  		//cout << "dcell n_pMHC: " << dcell->n_pMHC << endl;
	  		//cout << "ldc matured mdc; mdc n_pMHC: " << dcell->n_pMHC << endl;
	  		return D_MATURE;
	  }
	  return D_IMMATURE;
  } // end if D_IMMATURE
  
  if (current_state == D_MATURE) 
  { 
	  if (counter - dcell->time_in >= p.MAX_TIME_D_MATURE - dcell->birthtime )   
	  {       
	        // release tcells
	  	  for (int x = 0; x < dcell->tbind.size(); x++) { 	  
	  		for (int i = 1; i <= 8; i++) {
	  		  rrow = i / 3; if (rrow == 2) rrow = -1;
	  		  rcol = i % 3; if (rcol == 2) rcol = -1;
	  		  if (row + rrow >= 0 && row + rrow < MROWS 
	  					&& col + rcol >= 0 && col + rcol < MCOLS) {
	  			for(int j = 0; j < microcompartment::NUM_CELLS; j++){	    
	  			  if (M[row + rrow][col + rcol].cell[j] != NULL) {
	  				if (dcell->tbind[x] == M[row + rrow][col + rcol].cell[j]->id) {
	  					((Tcell*)M[row + rrow][col + rcol].cell[j])->bound_to_dcell = 0;
	  					M[row + rrow][col + rcol].cell[j]->next_state = H_RESTING;
	  				}
	  			  }
	  			}	    
	  		  }
	  		}    
	  	  }
	  		  return D_DEAD;
	  	}
	  
    // look for a neighboring h effector. If found, dcell becomes 
    // licensed with probability  PROB_eHELPER_LICENSES_MDC
    
    bool found_effector = false;

    for(int i = 1; i <= 8 && !found_effector; i++) 
    {
      rrow = i / 3; if (rrow == 2) rrow = -1;
      rcol = i % 3; if (rcol == 2) rcol = -1;
      if (row + rrow >= 0 && row + rrow < MROWS && col + rcol >= 0 && col + rcol < MCOLS) 
      {
    	  for(int j = 0; j < microcompartment::NUM_CELLS; j++)
    	  {
    		  if (M[row + rrow][col + rcol].cell[j] != NULL &&
    				  M[row + rrow][col + rcol].cell[j]->is_hcell() )
    		  {
	    
    			  if (M[row + rrow][col + rcol].cell[j]->state  == H_EFFECTOR) 
    			  {	      
    				  found_effector = true;
    			  }
    		  }
    	  }
      }  
    }
    if (found_effector) 
    {
      if (random.Next(0,10000)/100.0 <= p.PROB_eHELPER_LICENSES_MDC()) 
      {
    	  dcell->birthtime = counter + (int)((random.Next(0,10000)/100.0 - 50) / 50.0 * p.DCELL_TIME_BEING_LICENSED);
    	  dcell->n_pMHC =  p.UNBINDING_THRESHOLD() + 150;
    	  dcell->initial_pMHC = dcell->n_pMHC;
    	  dcell->licensing_time = counter;
	  //cout << "dcell " << dcell->id << " became licensed; n_pMHC is: " << dcell->n_pMHC << endl;
    	  return D_LICENSED;
      }
      else
      {
    	  return D_MATURE;
      }
    }
    return D_MATURE;
  }

  if (current_state == D_LICENSED)
  {
    // check the time being licensed for the licensed dcell
    p.DCELL_TIME_BEING_LICENSED = dcell->normal (21600, 2400); // choose from a normal distribution (36 +/- 4 hours)
    if (counter - dcell->licensing_time > p.DCELL_TIME_BEING_LICENSED) 
    {
      // dcell dies and release bound tcells
      
      for (int x = 0; x < dcell->tbind.size(); x++) 
      {
    	for (int i = 1; i <= 8; i++) 
    	{
    	  rrow = i / 3; if (rrow == 2) rrow = -1;
    	  rcol = i % 3; if (rcol == 2) rcol = -1;
    	  if (row + rrow >= 0 && row + rrow < MROWS && col + rcol >= 0 && col + rcol < MCOLS) 
    	  {
    		for(int j = 0; j < microcompartment::NUM_CELLS; j++)
    		{
    		  if (M[row + rrow][col + rcol].cell[j] != NULL) 
    		  {
    		    if (dcell->tbind[x] == M[row + rrow][col + rcol].cell[j]->id) 
    		    {
    		      ((Tcell*)M[row + rrow][col + rcol].cell[j])->bound_to_dcell = 0;
    		      M[row + rrow][col + rcol].cell[j]->next_state = H_RESTING;
    		    }
    		  }
    		}
    	  }
    	}
      }
      return D_DEAD;
    }
    return D_LICENSED;
  }
  
} // end next_state_dcells

void LymphNode::move_dendriticcell()
{
  int* ind = new int[LD.size()];
  int ri = 0; // saves overhead to re-use variable

  for(int i = 0; i < LD.size(); i++){
    ind[i] = i;
  }
  for(int i = 0; i < LD.size(); i++){
    ri = random.Next(0, LD.size() - 1);
    swap(ind[i], ind[ri]);
  }

  for(int iter = 0; iter < LD.size(); iter++){
    cellpos* mac = LD.elem(ind[iter]);
    d_cell* pd = ((d_cell*)mac->cell_pointer); 

    move_dendriticcell_loop(mac, pd);

  }

  delete[] ind;

  for(int i = 0; i < NEFFERENT; i++){
    if(M[efferent[2 * i]][efferent[2 * i + 1]].has_dcell()){
      delete_dcell(efferent[2 * i], efferent[2 * i + 1]);
    }
  }
}

void LymphNode::move_dendriticcell_loop(cellpos* mac, d_cell* pd)
{
  int row = mac->row;
  int col = mac->col;
  int dcID = mac->cell_pointer->id;
  
  int directionRow = 0; // difference between direction of old row and new
  int directionCol = 0; // difference between direction of old col and new
  
  // first, pick a direction
  DIRECTIONS move = (DIRECTIONS)random.Next(1, 4);
  
  // Then, I find my attached tcell buddies that will move with me
  const int cellsToMoveSize = 5;
  int cellsToMove[cellsToMoveSize][cellsToMoveSize];
  int freeCellSpaces[cellsToMoveSize][cellsToMoveSize];
  for(int i = 0; i < cellsToMoveSize; i++){
    for(int j = 0; j < cellsToMoveSize; j++){
      cellsToMove[i][j] = 0;
      freeCellSpaces[i][j] = 0;
    }
  }
  
  for(int i = -2; i <= 2; i++){
    if(row + i >= MROWS){
      move = NONE;
      return;
    }else if(row + i < 0){
      move = NONE;
      return;
    }else{
      for(int j = -2; j <= 2; j++){
	if(col + j >= MCOLS){
	  move = NONE;
	  return;
	}else if(col + j < 0){
	  move = NONE;
	  return;
	}else{
	  for(int k = 0; k < microcompartment::NUM_CELLS; k++){
	    if(withinRowsAndCols(row + i, col + j) && M[row + i][col + j].has_dcell()){
	      if(M[row + i][col + j].cell[0]->id == dcID){
		cellsToMove[i + 2][j + 2] = microcompartment::NUM_CELLS;
	      }else{
		// all compartments full and none are moving
	      }
	    }else if(withinRowsAndCols(row + i, col + j) && M[row + i][col + j].cell[k] != NULL){
	      if(((Tcell*)M[row + i][col + j].cell[k])->bound_to_dcell == dcID){
		cellsToMove[i + 2][j + 2]++;
	      }	    
	    }else if(withinRowsAndCols(row + i, col + j) && M[row + i][col + j].cell[k] == NULL){
	      freeCellSpaces[i + 2][j + 2]++;
	    }else if(!withinRowsAndCols(row + i, col + j)){
	      move = NONE;
	      checkRowsAndCols(row + i, col + j, "from move_dendriticcell_loop");
	      return;
	    }
	  }	 
	}
      }
    }
  }
  
  // Then, I check if my tcell buddies and I can move in that direction
  switch(move) {
  case EAST : 
    directionRow = 0;
    directionCol = 1;
    break;
  case SOUTH : 
    directionRow = 1;
    directionCol = 0;
    break;
  case WEST :
    directionRow = 0;
    directionCol = -1;
    break;
  case NORTH :
    directionRow = -1;
    directionCol = 0;
    break;
  case NONE :
    return;
    break;
  default :
    cerr << "Invalid move. Exiting." << endl;
    exit(16);
    break;
  }
  
  for(int i = -1; i <= 1; i++){
    for(int j = -1; j <= 1; j++){
      if(cellsToMove[i + 2][j + 2] > (cellsToMove[i + 2 + directionRow][j + 2 + directionCol] 
				      + freeCellSpaces[i + 2 + directionRow][j + 2 + directionCol])){
	move = NONE;
	return; // there isn't enough space for the attached tcells to move
      }
    }
  }

  // Then we all move
  if(EAST == move){
    if(col + directionCol + 1 >= MCOLS){
      move = NONE;
      return;
    }
    // cerr << "Moving a dcell! col increase " << "row: " << row << " col: " << col << " pd: " << pd << " " << pd->id << " counter: " <<  counter << endl;
    move_dendriticcell_hlpr(dcID, row + 1, col + 1, row + 1 + directionRow, col + 1 + directionCol);
    move_dendriticcell_hlpr(dcID, row, col + 1, row + directionRow, col + 1 + directionCol);
    move_dendriticcell_hlpr(dcID, row - 1, col + 1, row - 1 + directionRow, col + 1 + directionCol);
    move_dendriticcell_hlpr(dcID, row + 1, col, row + 1 + directionRow, col + directionCol);
    M[mac->row][mac->col].shift_cells(0);
    M[row + directionRow][col + directionCol].cell[0] = (Cell*)pd;
    mac->row += directionRow;
    mac->col += directionCol;
    move_dendriticcell_hlpr(dcID, row - 1, col, row - 1 + directionRow, col + directionCol);
    move_dendriticcell_hlpr(dcID, row + 1, col - 1, row + 1 + directionRow, col - 1 + directionCol);
    move_dendriticcell_hlpr(dcID, row, col - 1, row + directionRow, col - 1 + directionCol);
    move_dendriticcell_hlpr(dcID, row - 1, col - 1, row - 1 + directionRow, col - 1 + directionCol);
  }else if(SOUTH == move){
    if(row + directionRow + 1 >= MROWS){
      move = NONE;
      return;
    }
    // cerr << "Moving a dcell! row increase " << "row: " << row << " col: " << col << " pd: " << pd << " " << pd->id <<  " counter: " << counter << endl;
    move_dendriticcell_hlpr(dcID, row + 1, col - 1, row + 1 + directionRow, col - 1 + directionCol);
    move_dendriticcell_hlpr(dcID, row + 1, col, row + 1 + directionRow, col + directionCol);
    move_dendriticcell_hlpr(dcID, row + 1, col + 1, row + 1 + directionRow, col + 1 + directionCol);
    move_dendriticcell_hlpr(dcID, row, col - 1, row + directionRow, col - 1 + directionCol);
    M[mac->row][mac->col].shift_cells(0);
    M[row + directionRow][col + directionCol].cell[0] = (Cell*)pd;
    mac->row += directionRow;
    mac->col += directionCol;
    move_dendriticcell_hlpr(dcID, row, col + 1, row + directionRow, col + 1 + directionCol);
    move_dendriticcell_hlpr(dcID, row - 1, col - 1, row - 1 + directionRow, col - 1 + directionCol);
    move_dendriticcell_hlpr(dcID, row - 1, col, row - 1 + directionRow, col + directionCol);
    move_dendriticcell_hlpr(dcID, row - 1, col + 1, row - 1 + directionRow, col + 1 + directionCol);
  }else if(WEST == move){
    if(0 == col - 2){
      move = NONE;
      return;
    }
    // cerr << "Moving a dcell! col decrease " << "row: " << row << " col: " << col << " pd: " << pd << " " << pd->id <<  " counter: " << counter << endl;
    move_dendriticcell_hlpr(dcID, row + 1, col - 1, row + 1 + directionRow, col - 1 + directionCol);
    move_dendriticcell_hlpr(dcID, row, col - 1, row + directionRow, col - 1 + directionCol);
    move_dendriticcell_hlpr(dcID, row - 1, col - 1, row - 1 + directionRow, col - 1 + directionCol);
    move_dendriticcell_hlpr(dcID, row - 1, col, row - 1 + directionRow, col + directionCol);
    M[mac->row][mac->col].shift_cells(0);
    M[row + directionRow][col + directionCol].cell[0] = (Cell*)pd;
    mac->row += directionRow;
    mac->col += directionCol;
    move_dendriticcell_hlpr(dcID, row + 1, col, row + 1 + directionRow, col + directionCol);
    move_dendriticcell_hlpr(dcID, row + 1, col + 1, row + 1 + directionRow, col + 1 + directionCol);
    move_dendriticcell_hlpr(dcID, row, col + 1, row + directionRow, col + 1 + directionCol);
    move_dendriticcell_hlpr(dcID, row - 1, col + 1, row - 1 + directionRow, col + 1 + directionCol);
  }else if(NORTH == move){
    if(0 == row - 2){
      move = NONE;
      return;
    }
    // cerr << "Moving a dcell! row decrease " << "row: " << row << " col: " << col << " pd: " << pd << " " << pd->id <<  " counter: " << counter << endl;
    move_dendriticcell_hlpr(dcID, row - 1, col - 1, row - 1 + directionRow, col - 1 + directionCol);
    move_dendriticcell_hlpr(dcID, row - 1, col, row - 1 + directionRow, col + directionCol);
    move_dendriticcell_hlpr(dcID, row - 1, col + 1, row - 1 + directionRow, col + 1 + directionCol);
    move_dendriticcell_hlpr(dcID, row, col + 1, row + directionRow, col + 1 + directionCol);
    M[mac->row][mac->col].shift_cells(0);
    M[row + directionRow][col + directionCol].cell[0] = (Cell*)pd;
    mac->row += directionRow;
    mac->col += directionCol;
    move_dendriticcell_hlpr(dcID, row, col - 1, row + directionRow, col - 1 + directionCol);
    move_dendriticcell_hlpr(dcID, row + 1, col - 1, row + 1 + directionRow, col - 1 + directionCol);
    move_dendriticcell_hlpr(dcID, row + 1, col, row + 1 + directionRow, col + directionCol);
    move_dendriticcell_hlpr(dcID, row + 1, col + 1, row + 1 + directionRow, col + 1 + directionCol);
  }else{ // NONE == move
    return;
  }
}

void LymphNode::move_dendriticcell_hlpr(const int& dcellID, int row, int col, 
					int newRow, int newCol)
{
  checkRowsAndCols(row, col, "from move_dendriticcell_hlpr.  Problem with row and col");
  checkRowsAndCols(newRow, newCol, "from move_dendriticcell_hlpr.  Problem with newRow and newCol");

  cellpos* mac = NULL;
  for(int i = microcompartment::NUM_CELLS - 1; i >= 0; i--){
    if(M[row][col].cell[i] != NULL && ((Tcell*)M[row][col].cell[i])->bound_to_dcell == dcellID){
      if(M[row][col].cell[i]->is_hcell()){
	int pos = LH.whereis(M[row][col].cell[i]);
	if(pos != -1 && M[newRow][newCol].free_site()){
	  mac = LH.elem(pos);
	  LH.elem(pos)->row = newRow;
	  LH.elem(pos)->col = newCol;
	  h_cell* tp = (h_cell*)(mac->cell_pointer);
	  M[row][col].shift_cells(i);
	  M[newRow][newCol].cell[ M[newRow][newCol].which_site() ] = (Cell*)tp;
	}else{
	  if(-1 == pos){
	    cerr << "Could not find hcell in vector of hcells." << endl;
	  }
	  cerr << "problem with dendritic cell movement.  hcell" << " can't move. " << M[newRow][newCol].how_many_cells() << endl;
	}
      }else if(M[row][col].cell[i]->is_kcell()){
	int pos = LK.whereis(M[row][col].cell[i]);
	if(pos != -1 && M[newRow][newCol].free_site()){
	  mac = LK.elem(pos);
	  LK.elem(pos)->row = newRow;
	  LK.elem(pos)->col = newCol;
	  k_cell* tp = (k_cell*)(mac->cell_pointer);
	  M[row][col].shift_cells(i);
	  M[newRow][newCol].cell[ M[newRow][newCol].which_site() ] = (Cell*)tp;
	}else{
	  if(-1 == pos){
	    cerr << "Could not find kcell in vector of kcells." << endl;
	  }
	  cerr << "problem with dendritic cell movement. kcell" << " can't move. " << M[newRow][newCol].how_many_cells() << endl;
	}
      }else{
	cerr << "Problem in move_dendriticcell_hlpr. Exiting. " << endl;
	exit(19);
      }
    }
  }
}


