/* Here are the functions of the class LymphNode that have to do with the helper Tcells. */

#include "LN.h"
#include <math.h>
using namespace std;

int n_bound_hcells = 0;
int n_active_hcells = 0;
int n_effector_hcells = 0;
// create helper T cell in M and in the set LH
void LymphNode::create_hcell(int row, int col, int cell_index, int h_time,
			     int cell_state, int division_counter, int cognate, int bCounter)
{
  int rand_num = 0;
  checkRowsAndCols(row, col, "from create_hcell");
  cellpos cp;
  
  cp.row = row;
  cp.col = col;

  h_cell* ph = NULL;
  try
  {
	ph = new h_cell(); 
  }
  catch (bad_alloc error)
  {
	  cout << error.what() << endl;
	  cout<< "Memory could not be allocated to new hcell " << endl;
	  cout << "Exiting...." << endl;
	  exit(1);
  }
  if (ph != NULL)
  {
  	index_t4 += 3;
	ph->birthtime = h_time;
	ph->id = index_t4; // c_time;
	ph->state = cell_state;
	ph->next_state = cell_state;
	ph->bound_to_dcell = 0;
	ph->division_counter = division_counter;
	ph->time_in = counter;
	ph->cognate = cognate;
	allFirstThousandTCells ? ph->inFirstThousandTCells = true, allFirstThousandTCells-- : ph->inFirstThousandTCells = false;
	ph->persistentPath = NONE;
	ph->persistentPathLength = 0;
	ph->birthRow = row;
	ph->birthCol = col;
       // decide if the hcell is cognate
       //rand_num = rnd.Next(1, p.COGNATE_HCELL);
       rand_num = random.Next(1, p.COGNATE_HCELL);
       if (rand_num == 1)
		ph->cognate = 0;
	else 
		ph->cognate = 1;	
	if(counter > 0){
	ph->birthRow = row;
	ph->birthCol = col;
	}
	cp.cell_pointer = ph; 
	
	LH.push_back(cp);
	M[row][col].cell[cell_index] = ph;
	
	// vector of cognate resting hcells
	if(ph->cognate == 0 && ph->state == H_RESTING)
	{
		vector <float> temp;
                temp.push_back(counter/600);
		temp.push_back(ph->id);
		temp.push_back(ph->binding_prob);
		temp.push_back(ph->bound);
		temp.push_back(ph->activation_prob);
		temp.push_back(ph->activated);
		temp.push_back(ph->activation_time);
		temp.push_back(ph->searchTime);
		temp.push_back(ph->transitTime);
		hCognate.push_back(temp);
		n_cog_hcells++;
	}
	n_hcells++;
} // if ph != NULL
}

void LymphNode::delete_hcell(int row, int col, int cell_index)
{
  microcompartment* ptr = &(M[row][col]);
  Cell* cp = ptr->cell[cell_index];
  ((Tcell*)cp)->dcContacted.clear();
  ((Tcell*)cp)->dcContactedActive.clear();

  if(int dcID = ((Tcell*)cp)->bound_to_dcell){
    int pos = LD.whereis(dcID);
    if(pos != -1){
      ((d_cell*)((LD.elem(pos)->cell_pointer)))->remove_tcell(cp->id);
    }else{
      cerr << "Can't find dcell " << dcID << endl;
    }
  }
  M[row][col].shift_cells(cell_index);
  LH.erase(cp);
  //delete cp;
}

// create initial hcells
void LymphNode::set_helpercells_random()
{
  int newrow = 0;
  int newcol = 0;
  int cognate = 0; // it saves overhead to re-use this variable
  int cell_index = 0; // it saves overhead to re-use this variable
  bool isAL = false;
	
  for(int i = 0; i < p.INITIAL_rHELPERT_CELLS(); i++){
    do
    {
       isAL = false;
	newrow = random.Next(0, MROWS-1);
       newcol = random.Next(0, MCOLS-1);
       for ( int j = 0; j < NAFFERENT ; j++)
       {
	   if ( (newrow == afferent[2*j]) && (newcol == afferent[2*j+1]))
	   	isAL = true;
	   
	}
    } while (isAL);
    if (M[newrow][newcol].free_site() ){
	  //cout << "p.COGNATE_HCELL is " << p.COGNATE_HCELL << endl;
      (i % p.COGNATE_HCELL == 0) ? cognate = 0: cognate = 1;	
      cell_index = M[newrow][newcol].which_site();
      //cout << "Calling create_hcell() from set_helpercells_random() " << endl;
      create_hcell(newrow, newcol, cell_index, -random.Next(p.MIN_H_CELL_AGE, p.MAX_H_CELL_AGE), H_RESTING, 0, cognate, counter);
    } // end if
  } // end while
  
  for(int i = 0; i < p.INITIAL_aHELPERT_CELLS(); i++){
    newrow = random.Next(0, MROWS-1);
    newcol = random.Next(0, MCOLS-1);
    
    if (M[newrow][newcol].free_site() ){
      (i % p.COGNATE_HCELL == 0) ? cognate = 0: cognate = 1;	
      cell_index = M[newrow][newcol].which_site();
      create_hcell(newrow, newcol, cell_index, -random.Next(p.MIN_H_CELL_AGE, p.MAX_H_CELL_AGE), H_ACTIVE, 0, cognate, counter);
    }
  }
  for(int i = 0; i < p.INITIAL_eHELPERT_CELLS(); i++){
    newrow = random.Next(0, MROWS-1);
    newcol = random.Next(0, MCOLS-1);

    if (M[newrow][newcol].free_site()){
      (i % p.COGNATE_HCELL == 0) ? cognate = 0: cognate = 1;	
      cell_index = M[newrow][newcol].which_site();
      create_hcell(newrow, newcol, cell_index, -random.Next(p.MIN_H_CELL_AGE, p.MAX_H_CELL_AGE), H_EFFECTOR, 0, cognate, counter);
    }
  }
}

void LymphNode:: recruit_h_cells() 
{
  int cell_index = 0; // it saves overhead to re-use this variable
  int cognate = 0; // it saves overhead to re-use this variable

  for (int i = 0; i < NSOURCES; i++) {
    if (M[sources[2 * i]][sources[2 * i + 1]].free_site()) {
      float prob_event = random.Next(0,10000)/100.0;
      // dcs are surrogate for pathogen
      if (prob_event <= p.PROB_H_RECRUITMENT + (float)(NMD + NLD) * p.PROB_H_ENHANCED_RECRUITMENT() ) {
	int random_birth = counter - random.Next(p.MIN_H_CELL_AGE, p.MAX_H_CELL_AGE);
	(NEW_H_CELLS % p.COGNATE_HCELL == 0) ? cognate = 0 : cognate = 1;
	cell_index = M[sources[2 * i]][sources[2 * i + 1]].which_site();
	create_hcell(sources[2 * i], sources[2 * i + 1], cell_index, random_birth, H_RESTING, 0, cognate, counter);
	NEW_H_CELLS++;
      }
    } // end if free_site
  } // end for
}	

void Tcell::add_dcell(long dcell_id)
{
  dcContacted.push_back(dcell_id);
}

void Tcell::remove_dcell(long dcell_id)
{
  for (int j = 0; j < dcContacted.size(); j++) {
    if (dcContacted[j] == dcell_id) {
      dcContacted.erase((vector <int>::iterator)&dcContacted[j]);
    }
  }
}

void Tcell::add_dcellActive(long dcell_id)
{
  dcContactedActive.push_back(dcell_id);
}
void Tcell::remove_dcellActive(long dcell_id)
{
  for (int j = 0; j < dcContacted.size(); j++) {
    if (dcContactedActive[j] == dcell_id) {
      dcContactedActive.erase((vector <int>::iterator)&dcContactedActive[j]);
    }
  }
}

/* This function is used to update the state of a helper T cell. This 
   routine defines the rules for H Tcells */
int LymphNode::next_state_helperTcell(cellpos* ptr, int nDC)  
{
  int current_state = ptr->cell_pointer->state;
  
  if (current_state == H_RESTING)  { 
   // if (nDC >= 77 )
     //   cout << "Calling next_state_helperTcell_resting() " << endl;
    return next_state_helperTcell_resting(ptr, nDC);
  } // end if RESTING

  if (current_state == H_NOTMOVE_BY_DC) {
    return next_state_helperTcell_notmove(ptr);
  }

  if (current_state == H_ACTIVE) { 
    return next_state_helperTcell_active(ptr);
  }

  if  (current_state == H_EFFECTOR) { 
    return next_state_helperTcell_effector(ptr);
  } // end H_EFFECTOR

  //if(current_state == H_BOUND_TO_IDC){
    //return next_state_helperTcell_bound(ptr);
  //} // H_BOUND_TO_IDC
  return H_DEAD; // function should not get here
} // end next_state_helperTcell

int LymphNode::next_state_helperTcell_resting(cellpos* ptr, int nDC)
{
  h_cell* hcell = (h_cell*)(ptr->cell_pointer);
  int row = ptr->row; // it saves time to re-use this variable
  int col = ptr->col; // it saves time to re-use this variable
  int rrow = 0;  // it saves time to re-use this variable
  int rcol = 0;  // it saves time to re-use this variable
  float a,b,x = 0; // to calculate binding and activation probabilities
  p.setPROB_BINDING_T4_WITH_DC(0);

  
  // check for that the cell age is is range, otherwise kill it.
  if (counter - hcell->birthtime >= p.MAX_H_CELL_AGE) {
    return H_DEAD;
  }
  // look for a neighbor dcell 

  for(int i = 1; i <= 8; i++) {
    // look around at our neighbors
    rrow = i / 3; if (rrow == 2) rrow = -1;
    rcol = i % 3; if (rcol == 2) rcol = -1;   
   
     //  cout << "Checking neighbors of this hTcell to find DC" << endl;
    if (row + rrow >= 1 && row + rrow < MROWS - 1 && col + rcol >= 1 && col + rcol < MCOLS - 1){

      // if a neighbor dcell has less than 
      // the maximum tcells bound allowed, then bound to the dcell with 
      // probability PROB_BINDING_T4_WITH_DC
      
      // hit counter 
      if (M[row + rrow][col + rcol].has_dcell() && M[row + rrow][col + rcol].cell[0]->state != D_IMMATURE) { 

//	   cout << "Found a neighbor MDC/LDC" << endl;
    	 bool addCell = true;	
	// count the number of times a tcell touches any dcell
	     d_cell* dcell = (d_cell*)M[row+rrow][col+rcol].cell[0];
	     int dcLoc = dcell->id / 3 - 1;
	     dcell->n_hcell_contacts++;
	     
	     if (dcell->state == D_MATURE)
	     {
	    	 hMDCContactsTotal[dcLoc][0] = dcell->id; // can move somewhere else so it is only done once
	    	 hMDCContactsTotal[dcLoc][1] = dcell->initial_pMHC;
	    	 hMDCContactsTotal[dcLoc][2] = dcell->state;	    	 
	    	 hMDCContactsTotal[dcLoc][3]++;
	    	 hMDCContactsNaive[dcLoc]++;
	     }
	     if (dcell->state == D_LICENSED)
	     {
	     	 hLDCContactsTotal[dcLoc][0] = dcell->id; // can move somewhere else so it is only done once
	     	 hLDCContactsTotal[dcLoc][1] = dcell->initial_pMHC;
	     	 hLDCContactsTotal[dcLoc][2] = dcell->state;	     	 
	     	 hLDCContactsTotal[dcLoc][3]++;
	     	 hLDCContactsNaive[dcLoc]++;
	     }
	     
	     hcell->n_DcContacts++;

	     if(hcell->cognate == dcell->cognate){
           //cout << "Incrementing hContactsCognate[dcLoc] " << endl;
	    	 if (dcell->state == D_MATURE)
	    	 {
	    		 hMDCContactsCognate[dcLoc]++;
	    		 hMDCContactsNaiveCognate[dcLoc]++;
	    	 }
	    	 if (dcell->state == D_LICENSED)
	    	 {
	    	 	 hLDCContactsCognate[dcLoc]++;
	    	 	 hLDCContactsNaiveCognate[dcLoc]++;
	    	 }
	    	 
	     }
	     else{
	  // count the number of times a non-cognate hcell touches the dc
	    	 if (dcell->state == D_MATURE)
	    		 hMDCContactsNonCognate[dcLoc]++;
	    	 if (dcell->state == D_LICENSED)
	    		 hLDCContactsNonCognate[dcLoc]++;
	     }

	// check if the dcell is a first DC contact
	     if (!hcell->dcContacted.size()){ // hcell has not contacted a DC yet
	    	 if (hcell->time_in >= 0) { 
	    		 if ((counter - hcell->time_in) / 10 < 1000){ // make sure we don't try to access beyond the end of the vector
	    			 searchTime[(counter - hcell->time_in)/10]++; // search time in minutes
	    		 }
		        hcell->searchTime = counter - hcell->time_in;
		 	 for (int i = 0; i < hCognate.size(); i ++ )
			 {
			 	if (hCognate[i][0] == hcell->id)
			    		 hCognate[i][6] = hcell->searchTime;              
	    		}		
	    	 }
	     }
	     else  // if repeated contact, check if the contact is unique 
	     {
	    	 for(int x = hcell->dcContacted.size() - 1; x >= 0; x--){
	    		 if(hcell->dcContacted[x] == dcell->id){
	    			 addCell = false;
	    			 break;
	    		 }
	    	 } // end for
	     }
	if(addCell){  // if contact is unique
	  hcell->add_dcell(dcell->id);
	  dcell->add_hcell(hcell->id);
	  dcell->n_unique_hcell_contacts++;	  
	  hcell->n_uniqueDcContacts++;
	  if (dcell->state == D_MATURE)
		  hMDCContactsUnique[dcLoc]++;
	  if (dcell->state == D_LICENSED)
	  	  hLDCContactsUnique[dcLoc]++;
	}
	if (hcell->cognate == 0)
		hCogContactsFile << counter << "\t" << hcell->id << "\t" << ptr->row <<"," << ptr->col << "\t"
	         << hcell->n_DcContacts << "\t" << hcell->n_uniqueDcContacts << "\t" << dcell->id << "\t" << row+rrow << "," 
	         << col+rcol << "\t" << dcell->state << "\t" << dcell->tbind.size() << "\t";
	// end hit counter 
	// add the dcell to the vector, if the size of the vector is greater than 0
	if (dcell->tbind.size() < p.N_TCELL_BIND_DC) {
	  if(hcell->cognate == dcell->cognate){
		a = p.pMHC_FOR_MEDIAN_BINDING_PROB_HCELL();
		b = p.SHAPE_BINDING_PROB_CURVE_HCELL();
		x = dcell->n_pMHC;
		
		p.setPROB_BINDING_T4_WITH_DC(pow(1 + exp( -(x-a)/b) , -1) * 100); // logistic function
		//cout << "hcell id: " << hcell->id << ", ";
		//cout << "dcell id: " << dcell->id << ", ";
	       	//cout << "dcell n_pMHC: " << dcell->n_pMHC << endl;
		//cout << "PROB_BINDING_T4_WITH_DC: " << p.PROB_BINDING_T4_WITH_DC() << endl;
		hcell ->binding_prob =  p.PROB_BINDING_T4_WITH_DC(); // no contact history
		
		// add accumulated contact history
		//hcell ->binding_prob = hcell->contact_history + p.PROB_BINDING_T4_WITH_DC();
		
		hCogContactsFile << hcell->binding_prob << "\t";
		for (int i = 0; i < hCognate.size(); i ++ )
		{
			 if (hCognate[i][0] == hcell->id)
			     hCognate[i][1] = hcell->binding_prob;              
	    }	
	    if (random.Next(0,10000)/100.0 < hcell->binding_prob) {
	      // bind to D
	      //cout << "PROB_BINDING_T4_WITH_DC: " << p.PROB_BINDING_T4_WITH_DC() << endl;
	      //cout << "hcell " << hcell->id << " is bound to dcell " << dcell->id << endl;
	      dcell->add_tcell(hcell->id);  
	      hcell->bound_to_dcell = dcell->id;
	      int dcloc = dcell->id/3-1;
	      hcell->bound = 1;
	      for (int i = 0; i < hCognate.size(); i ++ )
	      {
	      	if (hCognate[i][0] == hcell->id)
	      		hCognate[i][2] = hcell->bound;              
	      }
	      hCogContactsFile << "yes" << "\t" << hcell->binding_history << endl;
	      if (dcell->state == D_MATURE)
	    	  hMDCBoundTotal[dcloc]++;
	      if (dcell->state == D_LICENSED)
	      	  hLDCBoundTotal[dcloc]++;
	       	      
	      // determine how long this tcell is to stay bound to a dcell
	      // tcell stays bound for hcell->not_move hours
	      hcell->time_bound = counter;       
          //cout << "Total number of hcell bound: " << ++n_bound_hcells << endl;
              	      
	      return H_NOTMOVE_BY_DC;
	    } // if prob
	    else  // hcell did not bind
	    {
		// accumulate contact history
	    	//hcell->contact_history = hcell->contact_history + p.PROB_BINDING_T4_WITH_DC();
	    	hCogContactsFile << "no" << "\t" << hcell->binding_history << endl;
	    }
	  } // end if cognate 
	} // if space available on dcell
	else  // no space to bind the hcell
		if (hcell->cognate == 0)
		    hCogContactsFile << hcell->binding_prob << "\t" << "no" << "\t" << hcell->binding_history << endl;
   } // if dcell
  } // inside boundaries
 }  // for 
  // reset contact history to 0 for no contact history version
  //hcell->contact_history = 0;
  return H_RESTING;       
}

int LymphNode::next_state_helperTcell_notmove(cellpos* ptr)
{
  h_cell* hcell = (h_cell*)(ptr->cell_pointer);
  d_cell* dcell = NULL;
  int DCpos = -1;
  int HCpos = -1;
  float prob_event;
  bool released = false;
  float a,b,x = 0; // to calculate binding and activation probabilities
  p.setPROB_DC_ACTIVATES_T4(0);
  // check for that the cell age is is range, otherwise kill it.
  if (counter - hcell->birthtime >= p.MAX_H_CELL_AGE) {
    return H_DEAD;
  }
  
    // locate dcell bound to the hcell
    
  if(hcell->bound_to_dcell){ // if dcell disappeared, hcell would still have this state, but not be bound to a dcell
      DCpos = LD.whereis(hcell->bound_to_dcell);
      
      if(-1 != DCpos){
    	  dcell = (d_cell*)(LD.elem(DCpos)->cell_pointer);
    	  HCpos = LH.whereis(hcell->id);
      }
      else{
          	cerr << counter << " Could not locate hcell " << hcell->id << "'s attached dcell " << hcell->bound_to_dcell << endl;
          }
  } 
  //cout << "DCpos is " << DCpos << endl;
  if (!dcell)
	cout << "dcell is NULL" << endl;
  
  if (dcell != NULL)
  {
	
  if (counter % 600 == 0) //  check status every hour
    {
	  hcell ->binding_history =  hcell ->binding_history +(dcell ->n_pMHC / (pow(pMHC_decay, 15))); // average n_pMHC of the DC during the hour
	  // release if pMHC is lower than unbinding threshold or binding history is strong enough
  	  if ((dcell ->n_pMHC < p.UNBINDING_THRESHOLD()) ||(hcell->binding_history >= p.UNBINDING_THRESHOLD() * (p.TIME_FOR_MEDIAN_ACTIVATION_PROB_HCELL()/600 + 1)))
  	  { 
  		if(-1 != HCpos){
  		   dcell->remove_tcell(hcell->id);
		   released = true;
		   //cout << "Released hcell " <<  hcell->id << " " << "dcell pMHC: " << dcell ->n_pMHC << " " << "hcell binding history: " << hcell->binding_history
		   // << endl;
		   hcell->bound_to_dcell = 0;
  		}
  		else{
  		    cerr << counter << " Could not locate hcell " << hcell->id << endl;
  		} 
  	  }
    }
      // time for binding is over, delete hcell from the dcell list of 
      // tcells bound to it.
  if ( counter > hcell->time_bound + hcell->not_move ) { 
    if(-1 != HCpos){
    	dcell->remove_tcell(hcell->id);
	released = true;
	//cout  << "Released hcell " <<  hcell->id << " after binding for 12 hours" << endl;
        hcell->bound_to_dcell = 0;				                      
    }
    else{
    	cerr << counter << " Could not locate hcell " << hcell->id << endl;
    }
  }
  } // if dcell != NULL	 
  if (released)
  {
    // Now the hcell has been released, check to see if it becomes active. 
    // Once released, hcells become active with probability p, with p having a 
    // linear distribution that is maximum (PROB_DC_ACTIVATES_T4) when there 
    // are no other tcells attached to the dcell, and is minimum (0), when 
    // dcell has all its possible bonds with hcells occupied
    prob_event = random.Next(0,10000)/100.0;
    a = p.TIME_FOR_MEDIAN_ACTIVATION_PROB_HCELL() / 600; // convert to hours
    b = p.SHAPE_ACTIVATION_PROB_CURVE_HCELL() / 10;
    x = hcell->binding_history / p.UNBINDING_THRESHOLD();
    //cout <<"a: " << a << " b: " << b << " binding_history: " << hcell->binding_history << " unbinding threshold: " <<  p.UNBINDING_THRESHOLD() << endl;
    //cout << "exp (-(x-a)/b): " << exp (-(x-a)/b) << endl;
    p.setPROB_DC_ACTIVATES_T4(pow(1 + exp (-(x-a)/b) , -1) * 100); // logistic function
    //cout <<"PROB_DC_ACTIVATES_T4: " << p.PROB_DC_ACTIVATES_T4() << endl;
    hcell->activation_prob = p.PROB_DC_ACTIVATES_T4();
    for (int i = 0; i < hCognate.size(); i ++ )
    {
    	if (hCognate[i][0] == hcell->id)
    	{
    	   hCognate[i][3] = hcell->activation_prob;
        }
    }	  

    if (prob_event <= p.PROB_DC_ACTIVATES_T4()) 
    {
    	
      hcell->birthtime = counter; 
      
      hcell->activation_time = counter - hcell->time_in;
      //vector< vector<float> >::iterator iter_ii;
      hcell->activated = 1;
      for (int i = 0; i < hCognate.size(); i ++ )
      {
    	  if (hCognate[i][0] == hcell->id)
    	  {
    		  hCognate[i][4] = hcell->activated;
    		  hCognate[i][5] = hcell->activation_time;
          }
      }	  
      //cout << "Activated hcell " <<  hcell->id << endl;
      //cout << "Total number of active hcells: " << ++n_active_hcells << endl;
      int dcloc = dcell->id/3-1;
      if (dcell->state == D_MATURE)
    	  hMDCActivatedTotal[dcloc]++;
      if (dcell->state == D_LICENSED)
          hLDCActivatedTotal[dcloc]++;
      
      return H_ACTIVE;
    }
    //cout << "hcell could not be activated" << endl;
    hcell->binding_history = 0;  // reset binding history to 0
    
    return H_RESTING;
  }
   
  return H_NOTMOVE_BY_DC;
}

int LymphNode::next_state_helperTcell_active(cellpos* ptr)
{
  int cell_index = 0;
  int row = ptr->row;
  int col = ptr->col;
  h_cell* hcell = (h_cell*)(ptr->cell_pointer);
  int rrow;
  int rcol;
  int current_state = ptr->cell_pointer->state;
  float prob_event;

  // check the active hcell age
  if (counter - hcell->birthtime >= p.HCELL_TIME_BEING_ACTIVE){
    return H_DEAD;
  }
  
  for(int i = 1; i <= 8; i++)
  {
    rrow=i/3; if(rrow == 2) rrow=-1;
    rcol=i%3; if(rcol==2)rcol=-1;
    
    if(withinRowsAndCols(row + rrow + 1, col + rcol + 1) && 
       withinRowsAndCols(row + rrow - 1, col + rcol - 1))
    {
      // hit counter
      if(M[row + rrow][col + rcol].has_dcell() &&
	 					M[row + rrow][col + rcol].cell[0]->state != D_IMMATURE ) 
      { 
	// count the number of times a tcell touches any non-immature dcell
    	  d_cell* dcell = (d_cell*)M[row + rrow][col + rcol].cell[0];
    	  int dcId = -1; // initialized to an invalid value
	// so we don't count one DC more than once in one time step
    	  if(dcell->id != dcId)
    	  {
    		  bool addCell = true;
    		  dcId = dcell->id;
    		  dcell->n_hcell_contactsActive++;
    		  if (dcell->state == D_MATURE)
    		  {
    			  hMDCContactsTotal[(dcell->id/3)-1][3]++;
    			  hMDCContactsTotal[(dcell->id/3)-1][0] = dcell->id;
    			  hMDCContactsTotal[(dcell->id/3)-1][1] = dcell->initial_pMHC;
    			  hMDCContactsTotal[(dcell->id/3)-1][2] = dcell->state;
    		  }
    		  if (dcell->state == D_LICENSED)
    		  {
    		      hLDCContactsTotal[(dcell->id/3)-1][3]++;
    		      hLDCContactsTotal[(dcell->id/3)-1][0] = dcell->id;
    		      hLDCContactsTotal[(dcell->id/3)-1][1] = dcell->initial_pMHC;
    		      hLDCContactsTotal[(dcell->id/3)-1][2] = dcell->state;
    		  }
    		     		  
    		  if(hcell->cognate == dcell->cognate)
    		  {
    			  if (dcell->state == D_MATURE)
    				  hMDCContactsCognate[(dcell->id/3)-1]++;
    			  if (dcell->state == D_LICENSED)
    				  hLDCContactsCognate[(dcell->id/3)-1]++;
    		  }
    		  else{
	    //count the number of times a non-cognate hcell touches the dc
    			  if (dcell->state == D_MATURE)
    				  hMDCContactsNonCognate[(dcell->id/3)-1]++;
    			  if (dcell->state == D_LICENSED)
    			      hLDCContactsNonCognate[(dcell->id/3)-1]++;
    		  }
	  // add the dcell to the vector, if the size of the vector is greater than 0,
	  // check if the dcell is a unique contact
    		  if(!hcell->dcContactedActive.size())
    		  {
    			  for(int x = 0; x < hcell->dcContactedActive.size(); x++)
    			  {
    				  if(hcell->dcContactedActive[x] == dcId)
    				  {
    					  addCell = false;
    					  break;
    				  }
    			  }
    		  }
    		  if(addCell)
    		  {
    			  hcell->add_dcellActive(((d_cell*)M[row + rrow][col + rcol].cell[0])->id);
    			  dcell->n_unique_hcell_contactsActive++;
    			  if (dcell->state == D_MATURE)
    				  hMDCContactsActiveUnique[(dcell->id/3)-1]++;
    			  if (dcell->state == D_LICENSED)
    			      hLDCContactsActiveUnique[(dcell->id/3)-1]++;
    		  }
    	  }
      } // end hit counter
      /*
      if(M[row + rrow][col + rcol].has_dcell() && // redundant
    		  M[row + rrow][col + rcol].cell[0]->state == D_IMMATURE)
      {
    	  if(hcell->cognate == ((d_cell*)(M[row + rrow][col + rcol].cell[0]))->cognate)
    	  {	  
    		  float prob_event = random.Next(0, 10000)/100.0;
    		  if(prob_event <= p.PROB_BINDING_PRIMED_T4_WITH_IDC())
    		  {
    			  int time_pT4_bound_to_IDC = random.Next(p.MIN_TIME_pTC_BOUND_TO_DC, p.MAX_TIME_pTC_BOUND_TO_DC);
    			  ((d_cell*)M[row + rrow][col + rcol].cell[0])->add_tcell(hcell->id);
    			  hcell->bound_to_dcell = M[row + rrow][col + rcol].cell[0]->id; // better way to say this
    			  hcell->not_move = time_pT4_bound_to_IDC;
    			  hcell->time_bound = counter;
    			  return H_BOUND_TO_IDC;
    		  }
    	  } // end if cognate
      } // end if has immature dcell
      */
    } // check neighbors
  } // end for
  
    // Active hcells have a division counter, to count the number of times a 
    // cell has divided. When the number of division reaches the maximum, the 
    // hcell becomes effector
  
  if ( hcell->division_counter >= p.N_DIVISIONS_T4) {
    //cout << "hcell " << hcell->id << " became an effector" << endl;
    //cout << "Total number of effector hcell: " << ++n_effector_hcells << endl;
    hcell->birthtime = counter; 
    return H_EFFECTOR;
  }
  
  // If the number of divisions is less than the maximum , an active hcell 
  // reproduces every aHCELL_DOUBLING_PERIOD. When a cell rproduces, an 
  // identical cell is created is the same, or a neighbor compartment. Both 
  // daughter cells increase the dision counter by 1
  else if (counter - hcell->birthtime >= p.aHCELL_DOUBLING_PERIOD ){ 
    // first randomize the order in which the neighbor compartments 
    // are searched
    
    int ind[8]; // can be in local place now, too
    for (int i=0;i<8;i++) ind[i]=i+1;
    for (int i=0;i<8;i++) {
      int ri = random.Next(0,7);
      swap(ind[i], ind[ri]);
    }
    
    bool replication_done = false;
    for(int i = 0; i < 8 &&  !replication_done; i++) {
      rrow = ind[i] / 3; if (rrow == 2) rrow=-1;
      rcol = ind[i] % 3; if (rcol == 2) rcol=-1;
      if (row+rrow>=0 && row+rrow<MROWS 
	  && col+rcol>=0 && col+rcol<MCOLS) {
	
	if (M[row+rrow][col+rcol].free_site()) {
	  // a neighbor compartment with free space is found
	  cell_index = M[row + rrow][col + rcol].which_site();
	  hcell->birthtime = counter;
	  hcell->division_counter++;
	  create_hcell(row + rrow, col + rcol, cell_index, counter, H_ACTIVE, hcell->division_counter, hcell->cognate, 0); // what does 0 stand for?
	  replication_done = true;
	  //cout << "hcell divided, total number of active hcells: " << ++n_active_hcells << endl;
	}
      }
    } // end for
    return H_ACTIVE;
  }
  return H_ACTIVE;
}

int LymphNode::next_state_helperTcell_effector(cellpos* ptr)
{
  int cell_index = 0;
  int row = ptr->row;
  int col = ptr->col;
  h_cell* hcell = (h_cell*)(ptr->cell_pointer);
  int rrow;
  int rcol;
  int current_state = ptr->cell_pointer->state;
  float prob_event;

  // check the age of effector hcells
  if (counter - hcell->birthtime >= p.MAX_TIME_H_EFFECTOR){
    return H_DEAD;
  }
  
  bool addCell = true;
  for(int i = 1; i <= 8; i++){
    rrow = i/3; if(rrow == 2) rrow = -1;
    rcol = i%3; if(rcol == 2) rcol = -1;
    
    if(withinRowsAndCols(row + rrow + 1, col + rcol + 1) && withinRowsAndCols(row + rrow - 1, col + rcol - 1)){
      // hit counter
      if(M[row + rrow][col + rcol].has_dcell() && M[row + rrow][col + rcol].cell[0]->state != D_IMMATURE)
      {
	// count the number of times a tcell touches any non-immature dcell
	d_cell* dcell = (d_cell*)M[row + rrow][col + rcol].cell[0];
	int dcId = -1; // initialized to invalid value; not necessary
	if(M[row + rrow][col + rcol].cell[0]->id != dcId){
	  dcId = M[row+rrow][col+rcol].cell[0]->id;
	  dcell->n_hcell_contactsActive++;
	  if (dcell->state == D_MATURE)
	  {
		  hMDCContactsTotal[(dcell->id/3)-1][3]++;
		  hMDCContactsTotal[(dcell->id/3)-1][0] = dcell->id; //should move this
		  hMDCContactsTotal[(dcell->id/3)-1][1] = dcell->initial_pMHC;
		  hMDCContactsTotal[(dcell->id/3)-1][2] = dcell->state;
	  }
	  if (dcell->state == D_LICENSED)
	  {
	  	  hLDCContactsTotal[(dcell->id/3)-1][3]++;
	  	  hLDCContactsTotal[(dcell->id/3)-1][0] = dcell->id; //should move this
	  	  hLDCContactsTotal[(dcell->id/3)-1][1] = dcell->initial_pMHC;
	  	  hLDCContactsTotal[(dcell->id/3)-1][2] = dcell->state;
	  }
	  	  
	  if(hcell->cognate == dcell->cognate){
		  if (dcell->state == D_MATURE)
			  hMDCContactsCognate[(dcell->id/3)-1]++;
		  if (dcell->state == D_LICENSED)
			  hLDCContactsCognate[(dcell->id/3)-1]++;
	  }else{
	    // count the number of times a non-cognate hcell touches the dc
		  if (dcell->state == D_MATURE)
			  hMDCContactsNonCognate[(dcell->id/3)-1]++;
		  if (dcell->state == D_LICENSED)
			  hLDCContactsNonCognate[(dcell->id/3)-1]++;
	  } 
	  // add the dcell to the vector, if the size of the vector is greater than 0, 
	  // check if the dcell is a unique contact
	  if (!hcell->dcContactedActive.size()){ // should take this out
	    for(int x = 0; x < hcell->dcContactedActive.size(); x++){
	      if(hcell->dcContactedActive[x] == dcId){
		addCell = false;
		break;
	      }
	    } // end for
	  }
	  if(addCell){  
	    hcell->add_dcellActive(((d_cell*)M[row+rrow][col+rcol].cell[0])->id);
	    dcell->n_unique_hcell_contactsActive++;
	  }
	}
      } // end hit counter
      /*
      if (M[row+rrow][col+rcol].has_dcell() && // redundant
    		  M[row+rrow][col+rcol].cell[0]->state == D_IMMATURE)
      {
    	  if(hcell->cognate == ((d_cell*)(M[row+rrow][col+rcol].cell[0]))->cognate)
    	  {   
    		  prob_event = random.Next(0,10000)/100.0;
    		  if(prob_event <= p.PROB_BINDING_PRIMED_T4_WITH_IDC())
    		  {
    			  int time_pT4_bound_to_IDC = random.Next(p.MIN_TIME_pTC_BOUND_TO_DC, p.MAX_TIME_pTC_BOUND_TO_DC);
    			  ((d_cell*)M[row+rrow][col+rcol].cell[0])->add_tcell(hcell->id);  
    			  hcell->bound_to_dcell = M[row+rrow][col+rcol].cell[0]->id;   
    			  hcell->not_move = time_pT4_bound_to_IDC; 
    			  hcell->time_bound = counter;       
    			  return H_BOUND_TO_IDC;
    		  } // end if prob_event
    	  } // end if cognate
      } // end if has immature dcell
      */
    } // end if in range
  } // end for
  return H_EFFECTOR;
} // end H_EFFECTOR
/*
int LymphNode::next_state_helperTcell_bound(cellpos* ptr)
{
  int cell_index = 0;
  int row = ptr->row;
  int col = ptr->col;
  h_cell* hcell = (h_cell*)(ptr->cell_pointer);
  int rrow;
  int rcol;
  int current_state = ptr->cell_pointer->state;
  float prob_event;
  // bool PTC_matures_IDC = false;

  // check for that the cell age is is range, otherwise kill it.
  if (counter - hcell->birthtime >= p.MAX_H_CELL_AGE) {
    return H_DEAD;
  }

  // check to see if the time for being bound is up
  if(counter > hcell->time_bound + hcell->not_move){
    // locate dcells bound to, and delete hcell from the list of tcells bound to it
    bool released = false;
    int dcRow = -1;
    int dcCol = -1;
    for(int i = 0; i <= 8 && !released; i++){
      if(i == 4){
	continue;
      }
      rrow = i/3; rrow -= 1;
      rcol = i%3; rcol -= 1;
      if(withinRowsAndCols(row + rrow, col + rcol)){  // should look at value in vector, not on grid
	if(M[row + rrow][col + rcol].has_dcell()){
	  if(M[row + rrow][col + rcol].cell[0]->id == hcell->bound_to_dcell){
	    ((d_cell*)M[row + rrow][col + rcol].cell[0])->remove_tcell(hcell->id);
	    hcell->bound_to_dcell = 0;
	    released = true;
	    dcRow = rrow;
	    dcCol = rcol;
	  }
	}
      }
    } // end for
    if(released){
      // mature the IDC with PROB_PTC_MATURES_IDC
      //cout << "Inside the funciton LymphNode::next_state_helperTcell_bound() " << endl;
      float prob_event = random.Next(0, 10000) / 100.0;
      //cout << "prob_event is " << prob_event << endl;
      if(M[row + dcRow][col + dcCol].cell[0]->state == D_IMMATURE && prob_event <= p.PROB_PTC_MATURES_IDC()){
        //cout << "PTC_matures_IDC is " << PTC_matures_IDC << endl;
	d_cell* dc = (d_cell*)M[row + dcRow][col + dcCol].cell[0];
	cellpos* cPos = NULL;
	for(int a = 0; a < LD.size(); a++){  // should use whereis
	  if(((d_cell*)(LD.elem(a)->cell_pointer))->id == dc->id){
	    cPos = LD.elem(a);
	    break;
	  }
	} // end for
	M[row + dcRow][col + dcCol].cell[0]->state = D_MATURE;
	M[row + dcRow][col + dcCol].cell[0]->birthtime = counter;
      }
      // return the cell to its previous state
      hcell->not_move = 0;
      if(hcell->division_counter >= p.N_DIVISIONS_T4){
	return H_EFFECTOR;
      }else{
	return H_ACTIVE;
      }
    }
  }
  return H_BOUND_TO_IDC;
} // H_BOUND_TO_IDC
*/
// This function is to move all hcells which are in the same state. 
// Tcells move sequentially to avoid collisions, but the sequential order 
// is ramdomized every time this function is called
void LymphNode::move_helpercell()
{
  int jumprow = -1;
  int jumpcol = -1;
  cellpos* mac = NULL;
  h_cell* ph = NULL;
  int ri = -1;
  int vectorSize = LH.size(); // saves overhead to declare this variable
  int* ind = NULL;
  bool isAL = false;
  
  try
  {
	ind = new int[vectorSize];
  }
  catch (bad_alloc error)
  {
	  cout << error.what() << endl;
	  cout<< "Memory could not be allocated to to ptr ind in move_helpercell() " << endl;
	  cout << "Exiting...." << endl;
	  exit(1);
  }
  if ( ind != NULL)
  {

  for (int i = vectorSize - 1; i >= 0; i--){
    ind[i] = i;
  }

  // ramdomize the order in which tcells are moved 
  for (int i = vectorSize - 1; i >= 0; i--) {
    ri = random.Next(0, vectorSize - 1);
    swap(ind[i], ind[ri]);
  }

  // The position of hcells is updated in the LH vector, and the pointers in 
  // the matrix M are updated
  // cerr << "h " <<  LH.whereis(19382) << " " << counter << endl;
  for(int i = vectorSize - 1; i >= 0; i--){
    mac = LH.elem(ind[i]); 
    ph = (h_cell*)(mac->cell_pointer);
    
    // find the most probable position to move, and move there if 
    // there is space to move. Otherwise, the cell remains static
    if (ph->state == H_RESTING){
      do
      {
	   isAL = false;
      	   find_most_probable_row_col(mac, jumprow, jumpcol);
	   for ( int j = 0; j < NAFFERENT ; j++)
          {
	      if ( (jumprow == afferent[2*j]) && (jumpcol == afferent[2*j+1]))
	      {
		  isAL = true;
	      }
          }
       }while (isAL);
      // update position in vector and matrix pointers
      if (M[jumprow][jumpcol].free_site() && M[mac->row][mac->col].find_cell(ph) != -1) {
	M[mac->row][mac->col].shift_cells(M[mac->row][mac->col].find_cell(ph));
	M[jumprow][jumpcol].cell[ M[jumprow][jumpcol].which_site() ] = ph;
	mac->row = jumprow;
	mac->col = jumpcol;
      }
    }else if(ph->state == H_ACTIVE || ph->state == H_EFFECTOR){
      // to slow down the non-naive hCells, change the multiplier of step_tcell
      if(counter % (2 * step_tcell) == 0){
       do
       {  
           isAL = false;
	    find_most_probable_row_col(mac, jumprow, jumpcol); 
	    for ( int j = 0; j < NAFFERENT ; j++)
           {
	       if ( (jumprow == afferent[2*j]) && (jumpcol == afferent[2*j+1]))
	       {
		    isAL = true;
		}
	    }
       }while (isAL);
	// update position in vector and matrix pointers
	if (M[jumprow][jumpcol].free_site() && M[mac->row][mac->col].find_cell(ph) != -1) {
	  M[mac->row][mac->col].shift_cells( M[mac->row][mac->col].find_cell(ph) );

	  M[jumprow][jumpcol].cell[ M[jumprow][jumpcol].which_site() ] = ph;
	  mac->row = jumprow;
	  mac->col = jumpcol;
	}
      }
    } // end else if
  } // end for

  // Delete cells that landed on efferent sites
  for (int i = 0; i < NEFFERENT; i++) {
    microcompartment* mp = &(M[ efferent[2*i] ][ efferent[2*i+1] ]); 
    int limit = M[ efferent[2*i] ][ efferent[2*i+1] ].how_many_cells();

    for(int pcell_index = 0, j = 0; j < limit; j++){
      Cell* cp = mp->cell[pcell_index];
      if (cp->is_hcell() ) 
      {
	h_cell* hptr = (h_cell*) cp;
	hptr->transitTime = counter - hptr->time_in;
	for (int i = 0; i < hCognate.size(); i ++ )
	{
		if (hCognate[i][0] == hptr->id)
			hCognate[i][7] = hptr->transitTime;              
	}
	
		
	if (cp->state == H_RESTING){
	  n_t4_cells_leave++;
	}
	else if (cp->state == H_ACTIVE)
	{
	  n_t4a_cells_leave++;
	}
	else if (cp->state == H_EFFECTOR)
	{
	  n_t4e_cells_leave++;					  
	}
       if (n_t4e_cells_leave == 1)
           time_first_eCD4_leave = counter;
	// add to pathLengths
	if(((Tcell*)cp)->inFirstThousandTCells // within data set
	   && ((Tcell*)cp)->persistentPath != NONE // already counted
	   && ((Tcell*)cp)->persistentPathLength > 0 // already counted
	   && ((Tcell*)cp)->persistentPathLength < pathLengths.size()){ // within bounds of vector
	  pathLengths[((Tcell*)cp)->persistentPathLength]++; 
	}
	// how far did it move in its lifetime
	if(((Tcell*)cp)->inFirstThousandTCells){
	  lifetimeDisplacement[(int)((Tcell*)cp)->calcDist(efferent[2*i], efferent[2*i+1])]++;
	}

	// prints for transitTimes
	int timei = cp->time_in;

	delete_hcell(efferent[2*i], efferent[2*i+1], pcell_index);

	if (timei >= 0) {
	  if (counter > timei && (counter - timei) / 60 < 960){ 
	    // make sure we don't try to access beyond the end of the vector
	    histo[(counter - timei)/60]++;
	  }
	} // end if timei >= 0
      }else{
	pcell_index++;
      }
    }
  }
  } // if ind != NULL
  delete[] ind;
}


