/* These functions are for the killer Tcell class */

#include "LN.h"
#include <math.h>

using namespace std;

int n_bound_kcells = 0;
int n_active_kcells = 0;
int n_effector_kcells = 0;
int n_memory_kcells = 0;
// create killer T cell in M and in the set LK
void LymphNode::create_kcell(int row, int col, int cell_index, int k_time, 
			     int cell_state, int division_counter, int cognate, int bCounter)
{
  int rand_num = 0;
  checkRowsAndCols(row, col, "from create_kcell");

  cellpos cp;
  cp.row = row;
  cp.col = col;
  k_cell* pk = NULL;
  try
  {
	pk = new k_cell();
  }
  catch (bad_alloc error)
  {
	  cout << error.what() << endl;
	  cout<< "Memory could not be allocated to new kcell " << endl;
	  cout << "Exiting...." << endl;
	  exit(1);
  }
  if ( pk != NULL)
  {
	index_t8 += 3;
	pk->birthtime = k_time;
	pk->id = index_t8; //c_time;
	pk->state = cell_state;
	pk->next_state = cell_state;
	pk->division_counter = division_counter;
	pk->time_in = counter;
	pk->cognate = cognate;
	allFirstThousandTCells ? pk->inFirstThousandTCells = true, allFirstThousandTCells-- : pk->inFirstThousandTCells = false;
	pk->persistentPath = NONE;
	pk->persistentPathLength = 0;
	pk->birthRow = row;
	pk->birthCol = col;
	// decide if the hcell is cognate
	//rand_num = rnd.Next(1, p.COGNATE_KCELL);
        rand_num = random.Next(1, p.COGNATE_KCELL);
       if (rand_num == 1)
		pk->cognate = 0;
	else 
		pk->cognate = 1;

	
	cp.cell_pointer = pk;
		
	LK.push_back(cp);  
	M[row][col].cell[cell_index] = pk;
	
	// vector of cognate resting kcells
	if(pk->cognate == 0 && pk->state == K_RESTING)
	{
		vector <float> temp;
		temp.push_back(counter/600);
		temp.push_back(pk->id);
		temp.push_back(pk->binding_prob);
		temp.push_back(pk->bound);
		temp.push_back(pk->activation_prob);
		temp.push_back(pk->activated);
		temp.push_back(pk->activation_time);
		temp.push_back(pk->searchTime);
		temp.push_back(pk->transitTime);
		kCognate.push_back(temp);
		n_cog_kcells++;
	}
	n_kcells++;
 } // if pk != NULL
}   

void LymphNode::delete_kcell(int row,int col, int cell_index)
{
  microcompartment* ptr = &(M[row][col]);
  Cell* cp = ptr->cell[cell_index];
  ((Tcell*)cp)->dcContacted.clear();
  ((Tcell*)cp)->dcContactedActive.clear();

  // if the kcell is attached to a dcell, remove it from tbind
  if(int dcID = ((Tcell*)cp)->bound_to_dcell){
    int pos = LD.whereis(dcID);
    if(pos != -1){
      ((d_cell*)((LD.elem(pos)->cell_pointer)))->remove_tcell(cp->id);
    }else{
      cerr << "Can't find dcell " << dcID << endl;
    }
  }
  LK.erase(cp);
  //delete ptr;
  M[row][col].shift_cells(cell_index);
}

void LymphNode::set_killercells_random()
{
  int newrow = 0;
  int newcol = 0;
  int cell_index = 0; // it saves overhead to re-use this variable
  int cognate = 0; // it saves overhead to re-use this variable
  bool isAL = false;
  for(int i = 0; i < p.INITIAL_rKILLERT_CELLS(); i++){
    do  // find non AL location
    {
       isAL = false;
    	newrow = random.Next(0, MROWS-1);
   	newcol = random.Next(0, MCOLS-1);
    	for ( int j = 0; j < NAFFERENT ; j++)
    	{
		if ( (newrow == afferent[2*j]) && (newcol == afferent[2*j+1]))
		{
			isAL = true;
		}
        }
    }while (isAL);
    if (M[newrow][newcol].free_site() ){
      (i % p.COGNATE_KCELL == 0) ? cognate = 0 : cognate = 1;
      cell_index = M[newrow][newcol].which_site();
      create_kcell(newrow, newcol, cell_index, -random.Next(p.MIN_K_CELL_AGE, p.MAX_K_CELL_AGE), K_RESTING, 0, cognate, counter);
    }
  }
  
  for(int i = 0; i < p.INITIAL_aKILLERT_CELLS(); i++){
    newrow = random.Next(0,MROWS-1);
    newcol = random.Next(0,MCOLS-1);
    
    if (M[newrow][newcol].free_site()){
      (i % p.COGNATE_KCELL == 0) ? cognate = 0 : cognate = 1;
      cell_index = M[newrow][newcol].which_site();
      create_kcell(newrow, newcol, cell_index, -random.Next(p.MIN_K_CELL_AGE, p.MAX_K_CELL_AGE), K_ACTIVE, 0, cognate, counter);
    }
  }
  
  for(int i = 0; i < p.INITIAL_mKILLERT_CELLS(); i++){
    newrow = random.Next(0,MROWS-1);
    newcol = random.Next(0,MCOLS-1);
    
    if (M[newrow][newcol].free_site()){
      (i % p.COGNATE_KCELL == 0) ? cognate = 0 : cognate = 1;
      cell_index = M[newrow][newcol].which_site();
      create_kcell(newrow, newcol, cell_index, -random.Next(p.MIN_K_CELL_AGE, p.MAX_K_CELL_AGE), K_MEMORY, 0, cognate, counter);
    }
  }
}

void LymphNode::recruit_k_cells() 
{
  float prob_event;
  int random_birth; 
  int cell_index = 0; 
  int cognate = 0;
  
  for (int i = 0; i < NSOURCES; i++) {
    if (M[sources[2*i]][sources[2*i+1]].free_site()) {
      prob_event = random.Next(0,10000)/100.0;
      
      // dcs are surrogate for pathogen
      if (prob_event <= p.PROB_K_RECRUITMENT + (float)(NMD + NLD) * p.PROB_H_ENHANCED_RECRUITMENT() ){
	random_birth = counter - random.Next(p.MIN_K_CELL_AGE, p.MAX_K_CELL_AGE);
	(NEW_K_CELLS % p.COGNATE_KCELL == 0) ? cognate = 0 : cognate = 1;
	cell_index = M[sources[2 * i]][sources[2 * i + 1]].which_site();
	create_kcell(sources[2*i], sources[2*i+1], cell_index, random_birth, K_RESTING, 0, cognate, counter);
	NEW_K_CELLS++;
      }
    } // end free_site
  } // end for 
}

// function to update the killer tcells
int LymphNode::next_state_killerTcell(cellpos* ptr) 
{
  int current_state = ptr->cell_pointer->state;
  if (current_state == K_RESTING) {
    return next_state_killerTcell_resting(ptr);
  }else if (current_state == K_NOTMOVE_BY_LDC){
    return next_state_killerTcell_notmove(ptr);
  }else if (current_state == K_ACTIVE) {
    return next_state_killerTcell_active(ptr);
  }else if (current_state == K_EFFECTOR){
    return next_state_killerTcell_effector(ptr);
  }else if (current_state == K_MEMORY){
    return next_state_killerTcell_memory(ptr);
  }
  return K_DEAD; // function should not get here
}

int LymphNode::next_state_killerTcell_resting(cellpos* ptr)
{
  k_cell* kcell = (k_cell *)ptr->cell_pointer;
  int row = ptr->row; // it saves time to re-use this variable
  int col = ptr->col; // it saves time to re-use this variable
  int rrow = 0; // it saves time to re-use this variable
  int rcol = 0; // it saves time to re-use this variable
  float a,b,x = 0; // to calculate binding and activation probabilities
  p.setPROB_BINDING_T8_WITH_LDC(0);
   
  // check the age of the kcell  
  if (counter - kcell->birthtime >= p.MAX_K_CELL_AGE) {
    return K_DEAD;
  }
  // look for a neighbor licensed dcell. with less than the maximum tcell  
  // bonds allowed. If found, bound to it with probability  
  // PROB_BINDING_T8_WITH_LDC 
  
  for(int i = 1; i <= 8; i++){
    // look around at our neighbors
    rrow = i/3; if (rrow == 2) rrow = -1;
    rcol = i%3; if (rcol == 2) rcol = -1;   
    if (row + rrow >= 1 && row + rrow < MROWS - 1 && col + rcol >= 1 && col + rcol < MCOLS - 1) {
      // if a neighbor dcell is mature or licensed and the dcell has less than 
      // the maximum tcells bound allowed, then bound to the dcell with 
      // probability PROB_BINDING_T4_WITH_DC
      // hit counter 
      if (M[row + rrow][col + rcol].has_dcell() && M[row + rrow][col + rcol].cell[0]->state != D_IMMATURE) { 
					
	bool addCell = true;
	// count the number of times a tcell touches any non-immature dcell
	d_cell* dcell = (d_cell*)M[row + rrow][col + rcol].cell[0];
	// so we don't count one DC more than once in one time step
	int dcId = dcell->id;
	dcell->n_kcell_contacts++; 
	kcell->n_DcContacts++;
	if (dcell->state == D_MATURE)
	{
		kMDCContactsTotal[(dcId/3)-1][0] = dcId;
		kMDCContactsTotal[(dcId/3)-1][1] = dcell->initial_pMHC;
		kMDCContactsTotal[(dcId/3)-1][2] = dcell->state;
		kMDCContactsTotal[(dcId/3)-1][3]++;
		kMDCContactsNaive[(dcId/3)-1]++;
	}
	if (dcell->state == D_LICENSED)
	{
		kLDCContactsTotal[(dcId/3)-1][0] = dcId;
		kLDCContactsTotal[(dcId/3)-1][1] = dcell->initial_pMHC;
		kLDCContactsTotal[(dcId/3)-1][2] = dcell->state;
		kLDCContactsTotal[(dcId/3)-1][3]++;
		kLDCContactsNaive[(dcId/3)-1]++;
	}
	
	if(kcell->cognate == dcell->cognate){
		if (dcell->state == D_MATURE)
		{
			kMDCContactsCognate[(dcId/3)-1]++;
			kMDCContactsNaiveCognate[(dcId/3)-1]++;
		}
		if (dcell->state == D_LICENSED)
		{
			kLDCContactsCognate[(dcId/3)-1]++;
			kLDCContactsNaiveCognate[(dcId/3)-1]++;
		}
	}else{
	  // count the number of times a non-cognate hcell touches the dc
		if(dcell->state == D_MATURE)
			kMDCContactsNonCognate[(dcId/3)-1]++;
		if(dcell->state == D_LICENSED)
			kLDCContactsNonCognate[(dcId/3)-1]++;
	}

	// add the dcell to the vector, if the size of the vector is greater than 0, 
	// check if the dcell is a unique contact
	if (!kcell->dcContacted.size()){
		if (kcell->time_in >= 0) {
	       	if ((counter - kcell->time_in) / 10 < 1000)   // make sure we don't try to access beyond the end of the vector
	      			searchTime[(counter - kcell->time_in) / 10]++;   // search time in minutes
	   	        kcell->searchTime = counter - kcell->time_in;
		 	 for (int i = 0; i < kCognate.size(); i ++ )
			 {
			 	if (kCognate[i][0] == kcell->id)
			    		 kCognate[i][6] = kcell->searchTime;              
	    		}
	  }	
	}else{
	  for(int x = kcell->dcContacted.size() - 1; x >= 0; x--){
	    if(kcell->dcContacted[x] == dcId){
	      addCell = false;
	      break;
	    }
	  } // end for
	}
	if(addCell){  
	  kcell->add_dcell(dcell->id);
	  dcell->add_kcell(kcell->id);
	  dcell->n_unique_kcell_contacts++;
	  kcell ->n_uniqueDcContacts++;
	  if (dcell->state == D_MATURE)
		  kMDCContactsUnique[(dcId/3)-1]++;
	  if (dcell->state == D_LICENSED)
	  	  kLDCContactsUnique[(dcId/3)-1]++;
	} // end hit counter 
	
	if (kcell->cognate == 0)
		kCogContactsFile << counter << "\t" << kcell->id << "\t" << ptr->row <<"," << ptr->col << "\t"
		      << kcell->n_DcContacts << "\t" << kcell->n_uniqueDcContacts << "\t" << dcell->id << "\t" << row+rrow << "," 
		      << col+rcol << "\t" << dcell->state << "\t" << dcell->tbind.size() << "\t";
	if (dcell->state  == D_LICENSED  && dcell->tbind.size() < p.N_TCELL_BIND_DC) {
	  if(kcell->cognate == dcell->cognate){
		a = p.pMHC_FOR_MEDIAN_BINDING_PROB_KCELL();
		b = p.SHAPE_BINDING_PROB_CURVE_KCELL();
		x = dcell->n_pMHC;
		//cout << "ldcell n_pMHC: " << dcell->n_pMHC << endl;
		p.setPROB_BINDING_T8_WITH_LDC(pow(1 + exp(-(x-a)/b) , -1) * 100);

		//cout << "kcell id: " << kcell->id << " ";
		//cout << "ldcell id: " << dcell->id << " ";
		//cout << "ldcell n_pMHC: " << dcell->n_pMHC << endl;
		//cout << "PROB_BINDING_T8_WITH_LDC: " << p.PROB_BINDING_T8_WITH_LDC() << endl;
	 kcell->binding_prob = p.PROB_BINDING_T8_WITH_LDC();
	 // add accumulated contact history
        // kcell->binding_prob = kcell->contact_history + p.PROB_BINDING_T8_WITH_LDC();
        kCogContactsFile << kcell->binding_prob << "\t";
        
	    for (int i = 0; i < kCognate.size(); i ++ )
	    {
	      	if (kCognate[i][0] == kcell->id)
	      	    kCognate[i][1] = kcell->binding_prob;              
	    }	     	
	    if (random.Next(0,10000)/100.0 < kcell->binding_prob) {  
	      //cout << "PROB_BINDING_T8_WITH_LDC: " << p.PROB_BINDING_T8_WITH_LDC() << endl;
	      //cout << "kcell " << kcell->id << " is bound to ldcell " << dcell->id << endl;
	      dcell->add_tcell(kcell->id);
	      kcell->bound_to_dcell = dcId;
	      int dcloc = dcell->id/3-1;
	      if (dcell->state == D_MATURE)
	    	  kMDCBoundTotal[dcloc]++; 
	      if (dcell->state == D_LICENSED)
	      	  kLDCBoundTotal[dcloc]++;
	      kcell->bound = 1;
	      for (int i = 0; i < kCognate.size(); i ++ )
	      {
	      	  if (kCognate[i][0] == kcell->id)
	      	     kCognate[i][2] = kcell->bound;              
	      }
	      kCogContactsFile << "yes" << "\t" << kcell->binding_history << endl;
	      //cout << "Total number of bound kcells: " << ++n_bound_kcells << endl;
	      // time this tcell is to stay bound to that dcell 
	      //kcell->not_move = 12 * 600; // tcell remains bound for max 12 hrs
	      kcell->time_bound = counter; 
	      return K_NOTMOVE_BY_LDC;
	    } // end if prob_event
	    else  // kcell did not bind to DC
	    {
		// accumulate binding probability as contact history
	    	//kcell->contact_history = kcell->contact_history + p.PROB_BINDING_T8_WITH_LDC();
	    	kCogContactsFile << "no" << "\t"<< kcell->binding_history << endl;
	    }
	  } // end if cognate
	} // D_LICENSED
	else   // dcell is not licensed
		if (kcell->cognate == 0)
		      kCogContactsFile << kcell->binding_prob << "\t" << "no" << "\t" << kcell->binding_history << endl;
   } // if dcell
  } // within range
 } // for
  // reset contact history to 0 for no contact history version
  //kcell->contact_history = 0;
  return K_RESTING;
}

int LymphNode::next_state_killerTcell_notmove(cellpos* ptr)
{
  k_cell* kcell = (k_cell *)ptr->cell_pointer; // it saves overhead to re-use this variable
  d_cell* dcell = NULL;
  int DCpos = -1;
  int KCpos = -1;
  float prob_event;
  int row = ptr->row; // it saves overhead to re-use this variable
  int col = ptr->col; // it saves overhead to re-use this variable
  int rrow = 0; // it saves overhead to re-use this variable
  int rcol = 0; // it saves overhead to re-use this variable
  bool released = false;
  float a,b,x = 0; // to calculate binding and activation probabilities
  p.setPROB_LDC_ACTIVATES_T8(0);
  
  // check for that the cell age is is range, otherwise kill it.
  if (counter - kcell->birthtime >= p.MAX_K_CELL_AGE) {
     return K_DEAD;
  }
  
  // locate dcell bound to the kcell
  if(kcell->bound_to_dcell){
      DCpos = LD.whereis(kcell->bound_to_dcell);
      if(-1 != DCpos){
	  dcell = (d_cell*)(LD.elem(DCpos)->cell_pointer);
          KCpos = LK.whereis(kcell->id);
      }
      else{
	  cerr << counter << " Could not locate kcell " << kcell->id << "'s attached dcell " << kcell->bound_to_dcell << endl;		                }
       }
  if (dcell != NULL)
  {
  if (counter % 600 == 0) //  check status every hour
  {
      kcell ->binding_history = kcell ->binding_history + (dcell ->n_pMHC / (pow(pMHC_decay, 15))); // average n_pMHC of the DC during the hour
      if ((dcell ->n_pMHC < p.UNBINDING_THRESHOLD()) ||(kcell->binding_history >= p.UNBINDING_THRESHOLD() * (p.TIME_FOR_MEDIAN_ACTIVATION_PROB_KCELL()/600 + 1)))
      {
	  if(-1 != KCpos){
	       dcell->remove_tcell(kcell->id);
	       released = true;
	       //cout << "Released kcell " <<  kcell->id << " dcell pMHC is " << dcell ->n_pMHC 
	            //<< " kcell binding history " << kcell->binding_history << endl;
	       kcell->bound_to_dcell = 0;
	  }
	  else{                    
	       cerr << counter << " Could not locate kcell " << kcell->id << endl;				           }  
      }	// end if dcell->n_pMHC  
  }   // end if counter

  // Check if the time for being bound is up   
  if (counter > kcell->time_bound + kcell->not_move) {
    // bound time is up. find a LDC to release from	 
    if(-1 != KCpos){
        dcell->remove_tcell(kcell->id);
        released = true;
	//cout  << "Released kcell after binding for 12 hours " <<  kcell->id << endl;
        kcell->bound_to_dcell = 0;
    }
    else{
	 cerr << counter << " Could not locate kcell " << kcell->id << endl;
    }
       
   // for(int i = 1; i <= 8; i++) {  
     // rrow = i / 3; if (rrow == 2) rrow = -1;
     // rcol = i % 3; if (rcol == 2) rcol = -1;
      
     // if (row + rrow >= 0 && row + rrow < MROWS && col + rcol >= 0 && col + rcol < MCOLS) { // can change this with whereis fn
       //	if (M[row + rrow][col + rcol].has_dcell() && M[row + rrow][col + rcol].cell[0]->id == kcell->bound_to_dcell ) {
	   // ((d_cell *)M[row+rrow][col+rcol].cell[0])->remove_tcell(kcell->id);
	   // break; // the kcell is now released
       //}
     // }
   // } // end for
  } // end if counter 
  } // if dcell != NULL
   if (released){ 
    // Now the kcell has been released. The kcell becomes active with 
    // probability p, which is maximum (PROB_LDC_ACTIVATES_T8) when there were 
    // no other tcells bound to the dcell, and minimum (0) when all tcell 
    // sites are occupied 
        prob_event = random.Next(0,10000)/100.0;
        a = p.TIME_FOR_MEDIAN_ACTIVATION_PROB_KCELL() / 600; // convert to hours
        b = p.SHAPE_ACTIVATION_PROB_CURVE_KCELL() / 10;
        x = kcell->binding_history / p.UNBINDING_THRESHOLD();
        //cout <<" a: " << a << " b: " << b << " binding_history: " << kcell->binding_history << " unbinding threshold: " <<  p.UNBINDING_THRESHOLD() << endl;
        //cout << "exp (-(x-a)/b): " << exp (-(x-a)/b) << endl;
        p.setPROB_LDC_ACTIVATES_T8(pow(1 + exp (-(x-a)/b) , -1) * 100); // logistic function
        //cout <<"PROB_LDC_ACTIVATES_T8 " << p.PROB_LDC_ACTIVATES_T8() << endl;
        kcell->activation_prob = p.PROB_LDC_ACTIVATES_T8();
        for (int i = 0; i < kCognate.size(); i ++ )
        {
            if (kCognate[i][0] == kcell->id)
            {
            	kCognate[i][3] = kcell->activation_prob;
            }
        }
        
        if ( prob_event <= p.PROB_LDC_ACTIVATES_T8()) {
             kcell->birthtime = counter;
             
             kcell->activation_time = counter - kcell->time_in;
             //vector< vector<float> >::iterator iter_ii;
             kcell->activated = 1;
             for (int i = 0; i < kCognate.size(); i ++ )
             {
            	 if (kCognate[i][0] == kcell->id)
            	 {
            		 kCognate[i][4] = kcell->activated;
            		 kCognate[i][5] = kcell->activation_time;
            	 }
             }		 
	     //cout << "kcell " << kcell->id << " is activated " << endl;
	     int dcloc = dcell->id/3-1;
	     if (dcell->state == D_MATURE)
	    	 kMDCActivatedTotal[dcloc]++;
	     if (dcell->state == D_LICENSED)
	     	 kLDCActivatedTotal[dcloc]++;
             
	     return K_ACTIVE;
	     //cout << "Total number of active kcells: " << ++n_active_kcells << endl;
        }
        kcell->binding_history = 0; // reset kcell binding history to 0
        return K_RESTING;
  }
  return K_NOTMOVE_BY_LDC;
}

int LymphNode::next_state_killerTcell_active(cellpos* ptr)
{
  k_cell* kcell = (k_cell*)ptr->cell_pointer; // it saves overhead to re-use this variable
  int row = ptr->row; // it saves overhead to re-use this variable
  int col = ptr->col; // it saves overhead to re-use this variable
  int rrow = 0; // it saves overhead to re-use this variable
  int rcol = 0; // it saves overhead to re-use this variable
  
  // check the active kcell age  
  if (counter - kcell->birthtime >= p.KCELL_TIME_BEING_ACTIVE){
    return K_DEAD;
  }

  for(int i = 1; i <= 8; i++) {
    // look around at our neighbors
    rrow = i / 3; if (rrow == 2) rrow = -1;
    rcol = i % 3; if (rcol == 2) rcol = -1;   
    
    if (row+rrow >= 1 && row + rrow < MROWS - 1 && col + rcol >= 1 && col + rcol < MCOLS - 1) {
      // hit counter 
      if (M[row + rrow][col + rcol].has_dcell() && M[row + rrow][col + rcol].cell[0]->state != D_IMMATURE) { 	
	bool addCell = true;	
	// count the number of times a tcell touches any non-immature dcell
	int dcVectorSize = kcell->dcContactedActive.size();  
	d_cell* dcell = (d_cell*)M[row + rrow][col + rcol].cell[0];
	// so we don't count one DC more than once in one time step
	//if(M[row + rrow][col + rcol].cell[0]->id != dcId){
	int dcId = M[row + rrow][col + rcol].cell[0]->id;
	dcell->n_kcell_contactsActive++;
	if (dcell->state == D_MATURE)
	{
		kMDCContactsTotal[(dcell->id/3)-1][3]++;
		kMDCContactsTotal[(dcell->id/3)-1][0] = dcell->id;
		kMDCContactsTotal[(dcell->id/3)-1][1] = dcell->initial_pMHC;
		kMDCContactsTotal[(dcell->id/3)-1][2] = dcell->state;
	}
	if (dcell->state == D_LICENSED)
	{
		kLDCContactsTotal[(dcell->id/3)-1][3]++;
		kLDCContactsTotal[(dcell->id/3)-1][0] = dcell->id;
		kLDCContactsTotal[(dcell->id/3)-1][1] = dcell->initial_pMHC;
		kLDCContactsTotal[(dcell->id/3)-1][2] = dcell->state;
	}
	
	if(kcell->cognate == dcell->cognate){
		if (dcell->state == D_MATURE)
			kMDCContactsCognate[(dcell->id/3)-1]++;
		if (dcell->state == D_LICENSED)
			kLDCContactsCognate[(dcell->id/3)-1]++;
	}else{
		if (dcell->state == D_MATURE)
			kMDCContactsNonCognate[(dcell->id/3)-1]++;
		if (dcell->state == D_LICENSED)
			kLDCContactsNonCognate[(dcell->id/3)-1]++;
	}
	// add the dcell to the vector, if the size of the vector is greater than 0, 
	// check if the dcell is a unique contact
	if (!dcVectorSize){
	  for(int x = 0; x < dcVectorSize; x++){
	    if(kcell->dcContactedActive[x] == dcId){
	      addCell = false;
	      break;
	    }
	  } // end for
	}
	if(addCell){  
	  kcell->add_dcellActive(((d_cell *)M[row+rrow][col+rcol].cell[0])->id);
	  dcell->n_unique_kcell_contactsActive++;
	  if (dcell->state == D_MATURE)
		  kMDCContactsActiveUnique[(dcell->id/3)-1]++;
	  if (dcell->state == D_LICENSED)
	  	 kLDCContactsActiveUnique[(dcell->id/3)-1]++;
	}
      } // end hit counter 
    } // end in range
  } // end for
  
  // Active kcells have a division counter, to count the number of times a
  // cell has divided. When the number of division reaches the maximum, the
  // kcell becomes effector
  if (kcell->division_counter >= p.N_DIVISIONS_T8) {
    kcell->birthtime = counter;
    if (random.Next(0,10000)/100.0 < p.PROB_aK_BECOMES_mK()){
      //cout << "kcell " << kcell->id << " became a memory cell" << endl;
      //cout << "Total number of memory kcells: " << ++n_memory_kcells << endl;
      return K_MEMORY;
    }else{
      //cout << "Total number of effector  kcells: " << ++n_effector_kcells << endl;
      //cout << "kcell " << kcell->id << " became an effector cell" << endl;
      return K_EFFECTOR;
    }
  }else if (counter - kcell->birthtime >= p.aKCELL_DOUBLING_PERIOD) {
    // active kcells start to divide. If the number of divisions is less than 
    // the maximum, each active kcell reproduces every  aKCELL_DOUBLING_PERIOD
    bool replication_done = false;
    
    int ind[9]; // for determining which neighbor to look at first
    // randomize the order in which neighbor compartments are searched
    for (int i = 0; i <= 8; i++){ 
      ind[i] = i;
    }
    for (int i = 0; i <= 8; i++) {
      int ri = random.Next(0, 7);
      swap(ind[i], ind[ri]);
    }
    // look for a neighbor compartment with free space and create an 
    // identical cell in it
    for(int i = 0; i <= 8 &&  !replication_done; i++) {
      rrow = ind[i] / 3; 
      if (rrow==2) rrow =- 1;
      rcol = ind[i] % 3; if (rcol==2) rcol =- 1;
      if (row+rrow>=0 && row+rrow<MROWS
	  && col+rcol>=0 && col+rcol<MCOLS) {
	if (M[row+rrow][col+rcol].free_site()) {
	  int cell_index = M[row + rrow][col + rcol].which_site();
	  // (M[row + rrow][col + rcol].is_empty()) ? cell_index = 0 : cell_index = 1;
	  kcell->birthtime = counter;
	  kcell->division_counter++;
	  create_kcell(row+rrow, col+rcol, cell_index, counter, K_ACTIVE, kcell->division_counter, kcell->cognate, 0);
	  replication_done = true;
	  //cout << "kcell divided, total number of active kcells: " << ++n_active_kcells << endl;
	} // if
      } // if 
    } // for
    return K_ACTIVE;
  }
  return K_ACTIVE;
}

int LymphNode::next_state_killerTcell_effector(cellpos* ptr)
{
  int rrow;
  int rcol;
  float prob_event;
  
  int row = ptr->row;
  int col = ptr->col;
  k_cell *kcell = (k_cell *)ptr->cell_pointer;

  if (counter-kcell->birthtime >= p.MAX_TIME_K_EFFECTOR){
    return K_DEAD;
  }
    
  for(int i = 1; i <= 8; i++) {
    // look around at our neighbors
    rrow=i/3; if (rrow==2) rrow=-1;
    rcol=i%3; if (rcol==2) rcol=-1;   
    
    if (row+rrow>=1 && row+rrow<MROWS-1 && col+rcol>=1 && col+rcol<MCOLS-1) {
      // hit counter 
      if (M[row+rrow][col+rcol].has_dcell() && M[row + rrow][col + rcol].cell[0]->state != D_IMMATURE) { 
	bool addCell = true;
	// count the number of times a tcell touches any non-immature dcell
	int dcVectorSize = kcell->dcContactedActive.size();  
	d_cell* dcell = (d_cell *)M[row+rrow][col+rcol].cell[0];
	// so we don't count one DC more than once in one time step
	int dcId = M[row+rrow][col+rcol].cell[0]->id;
	dcell-> n_kcell_contactsActive++;
	if (dcell->state == D_MATURE)
	{
		kMDCContactsTotal[(dcell->id/3)-1][3]++;
		kMDCContactsTotal[(dcell->id/3)-1][0] = dcell->id;
		kMDCContactsTotal[(dcell->id/3)-1][1] = dcell->initial_pMHC;
		kMDCContactsTotal[(dcell->id/3)-1][2] = dcell->state;
	}
	if (dcell->state == D_LICENSED)
	{
		kLDCContactsTotal[(dcell->id/3)-1][3]++;
		kLDCContactsTotal[(dcell->id/3)-1][0] = dcell->id;
		kLDCContactsTotal[(dcell->id/3)-1][1] = dcell->initial_pMHC;
		kLDCContactsTotal[(dcell->id/3)-1][2] = dcell->state;
	}
	if(kcell->cognate == dcell->cognate){
		if (dcell->state == D_MATURE)
			kMDCContactsCognate[(dcell->id/3)-1]++;
		if (dcell->state == D_LICENSED)
			kLDCContactsCognate[(dcell->id/3)-1]++;
	}else{
		if (dcell->state == D_MATURE)
			kMDCContactsNonCognate[(dcell->id/3)-1]++;
		if (dcell->state == D_LICENSED)
			kLDCContactsNonCognate[(dcell->id/3)-1]++;
	} 
	// add the dcell to the vector, if the size of the vector is greater than 0, 
	// check if the dcell is a unique contact
	if (!dcVectorSize){
	  for(int x =0; x<dcVectorSize; x++){
	    if(kcell->dcContactedActive[x] == dcId){
	      addCell = false;
	      break;
	    }
	  } // end for
	}
	if(addCell){  
	  kcell->add_dcellActive(((d_cell *)M[row+rrow][col+rcol].cell[0])->id);
	  dcell->n_unique_kcell_contactsActive++;
	}
      } // end hit counter 
    } // end if in range
  } // end for
  return K_EFFECTOR;
}
 
int LymphNode::next_state_killerTcell_memory(cellpos* ptr)
{ 
  int rrow;
  int rcol;
  float prob_event;
  
  int row = ptr->row;
  int col = ptr->col;
  int cell_index;
  k_cell *kcell = (k_cell *)ptr->cell_pointer;

  // mKcells do not die. they release cytokines    
  int dcId = -1;
  int dcVectorSize =0;
  bool addCell = true;
  d_cell *dcell;
  
  for(int i = 1; i <= 8; i++) {
    // look around at our neighbors
    rrow=i/3; if (rrow==2) rrow=-1;
    rcol=i%3; if (rcol==2) rcol=-1;   
    
    if (row+rrow>=1 && row+rrow<MROWS-1 && col+rcol>=1 && col+rcol<MCOLS-1) {
      // hit counter 
      if (M[row+rrow][col+rcol].has_dcell() && M[row + rrow][col + rcol].cell[0]->state != D_IMMATURE) { 
	// count the number of times a tcell touches any non-immature dcell
	dcVectorSize = kcell->dcContactedActive.size();  
	dcell = (d_cell *)M[row+rrow][col+rcol].cell[0];
	// so we don't count one DC more than once in one time step
	//if(M[row+rrow][col+rcol].cell[0]->id != dcId){
	dcId = M[row+rrow][col+rcol].cell[0]->id;
	dcell-> n_kcell_contactsActive++;
	if (dcell->state == D_MATURE){
		kMDCContactsTotal[(dcell->id/3)-1][3]++;
		kMDCContactsTotal[(dcell->id/3)-1][0] = dcell->id;
		kMDCContactsTotal[(dcell->id/3)-1][1] = dcell->initial_pMHC;
		kMDCContactsTotal[(dcell->id/3)-1][2] = dcell->state;
	}
	if (dcell->state == D_LICENSED){
		kLDCContactsTotal[(dcell->id/3)-1][3]++;
		kLDCContactsTotal[(dcell->id/3)-1][0] = dcell->id;
		kLDCContactsTotal[(dcell->id/3)-1][1] = dcell->initial_pMHC;
		kLDCContactsTotal[(dcell->id/3)-1][2] = dcell->state;
	}	
	if(kcell->cognate == dcell->cognate){
		if (dcell->state == D_MATURE)
			kMDCContactsCognate[(dcell->id/3)-1]++;
		if (dcell->state == D_LICENSED)
			kLDCContactsCognate[(dcell->id/3)-1]++;
	}else{
	  // count the number of times a non-cognate kcell touches the dc
		if (dcell->state == D_MATURE)
			kMDCContactsNonCognate[(dcell->id/3)-1]++;
		if (dcell->state == D_LICENSED)
			kLDCContactsNonCognate[(dcell->id/3)-1]++;
	} 
	// add the dcell to the vector, if the size of the vector is greater than 0, 
	// check if the dcell is a unique contact
	if (dcVectorSize){
	  for(int x = 0; x < dcVectorSize; x++){
	    if(kcell->dcContactedActive[x] == dcId){
	      addCell = false;
	      break;
	      }
	  } // end for
	}
	if(addCell){  
	  kcell->add_dcellActive(((d_cell *)M[row+rrow][col+rcol].cell[0])->id);
	  dcell->n_unique_kcell_contactsActive++;
	}
      } // end hit counter 
    } // end if in range
  } // end while
  return K_MEMORY;
}

// This function moves the killer cells which are in the same state. It 
// works exactly the same way as the function move_helpercells, but 
// updates the vector LK
void LymphNode::move_killercell()
{
  int jumprow = -1; // it saves overhead to re-use this variable
  int jumpcol = -1; // it saves overhead to re-use this variable
  cellpos* mac = NULL; // it saves overhead to re-use this variable
  k_cell* pk = NULL; // it saves overhead to re-ue this variable
  int ri = -1; // it saves overhead to re-use this variable
  int vectorSize = LK.size();
  int* ind = NULL;
  bool isAL = false;
  
  try
  {
	ind = new int[vectorSize];
  }
  catch (bad_alloc error)
  {
	  cout << error.what() << endl;
	  cout<< "Memory could not be allocated to new ptr ind in move_killercell() " << endl;
	  cout << "Exiting...." << endl;
	  exit(1);
  }
  if ( ind != NULL)
  {
  
  for (int i = vectorSize - 1; i >= 0; i--){
    ind[i] = i;
  }
  for (int i = vectorSize - 1; i >= 0; i--) {
    ri = random.Next(0, vectorSize - 1);
    swap(ind[i], ind[ri]);
  }
  for(int i = vectorSize - 1; i >= 0; i--){
    mac = LK.elem(ind[i]);
    pk = (k_cell *)(mac->cell_pointer);

    if  (pk->state != K_NOTMOVE_BY_LDC)
    {
	if (pk->state == K_RESTING)
       {
         do
         {  
	     isAL = false;
            find_most_probable_row_col(mac, jumprow, jumpcol);
            for ( int j = 0; j < NAFFERENT ; j++)
            {
	       if ( (jumprow == afferent[2*j]) && (jumpcol == afferent[2*j+1]))
	       {
		   isAL = true;
		}
            }
         }while (isAL);
            
      // update position in vector and matrix pointers
         if (M[jumprow][jumpcol].free_site() && M[mac->row][mac->col].find_cell(pk) != -1) 
	  {

		M[mac->row][mac->col].shift_cells( M[mac->row][mac->col].find_cell(pk) );

		M[jumprow][jumpcol].cell[ M[jumprow][jumpcol].which_site() ] = pk;

		mac->row = jumprow;
		mac->col = jumpcol;
         }
      }// if (pk->state == K_RESTING)
      else if (pk->state == K_ACTIVE || pk->state == K_EFFECTOR || pk->state == K_MEMORY)	
      {
       if(counter % (2 * step_tcell) == 0)
       {
         do
         {  
           isAL = false;
	    find_most_probable_row_col(mac, jumprow, jumpcol); 
	    for ( int j = 0; j < NAFFERENT ; j++)
           {
	       if ( (jumprow == afferent[2*j]) && (jumpcol == afferent[2*j+1]))
	       {
		    isAL = true;
		}
	    }
         }while (isAL);
	// update position in vector and matrix pointers
	  if (M[jumprow][jumpcol].free_site() && M[mac->row][mac->col].find_cell(pk) != -1) {
	    M[mac->row][mac->col].shift_cells( M[mac->row][mac->col].find_cell(pk) );

	    M[jumprow][jumpcol].cell[ M[jumprow][jumpcol].which_site() ] = pk;
	    mac->row = jumprow;
	    mac->col = jumpcol;
	  } // end if (M[..
      } // end if (couter %..
     } // end else if  
     else pk->direction = 0; // coudnt move
    } // if  (pk->state != K_NOTMOVE_BY_LDC)
  }
 
  // Delete cells that landed on efferent sites
  for (int i = 0; i < NEFFERENT; i++) {
    microcompartment* mp = &(M[ efferent[2*i] ][ efferent[2*i+1] ]); 
    int limit = M[ efferent[2*i] ][ efferent[2*i+1] ].how_many_cells();

    for(int pcell_index = 0, j = 0; j < limit; j++){
      Cell* cp = mp->cell[pcell_index]; 
      
      if (cp->is_kcell() ) 
      {
    	  k_cell* kptr = (k_cell*)cp;
	  kptr->transitTime = counter - kptr->time_in;
	  for (int i = 0; i < kCognate.size(); i ++ )
	  {
		if (kCognate[i][0] == kptr->id)
			kCognate[i][7] = kptr->transitTime;              
	  } 
    	  if (cp->state ==  K_RESTING) 
    	  {
    		  n_t8_cells_leave++;
    	  }
    	  else if (cp->state == K_ACTIVE)
    	  {       	
    		  n_t8a_cells_leave++; 
    	  }
    	  else if (cp->state == K_EFFECTOR) 
    	  {
    		  n_t8e_cells_leave++;
    	  }				
    	  else if (cp->state == K_MEMORY) 
    	  {
    		  n_t8m_cells_leave++;
    	  }
         if (n_t8e_cells_leave == 1 || n_t8m_cells_leave == 1)
             time_first_eCD8_leave = counter;

	// add to pathLengths
	if(((Tcell*)cp)->inFirstThousandTCells // within data set
	   && ((Tcell*)cp)->persistentPath != NONE // already counted
	   && ((Tcell*)cp)->persistentPathLength > 0 // already counted
	   && ((Tcell*)cp)->persistentPathLength < pathLengths.size()){ // within bounds of vector
	  pathLengths[((Tcell*)cp)->persistentPathLength]++; 
	}
      
	// how far did it move its lifetime
	if(((Tcell*)cp)->inFirstThousandTCells){
	  lifetimeDisplacement[(int)((Tcell*)cp)->calcDist(efferent[2*i], efferent[2*i+1])]++;
	}
	delete_kcell(efferent[2*i],efferent[2*i+1], pcell_index);
      }else{
	pcell_index++;
      }    
    }
  }
  } // end if ind != NULL
  delete[] ind;
}

